#!/usr/bin/env python3

"""DIEGO
"""

# Gero Doose gero@bioinf.uni-leipzig.de
import sys
import argparse
import operator
import math
from datetime import datetime
import itertools
import numpy as np
from scipy.stats import ranksums
from scipy.stats import nbinom
from scipy.cluster.hierarchy import linkage, dendrogram
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt


parser = argparse.ArgumentParser(description='DIEGO (DIfferential altErnative splicinG detectiOn) \n')
parser.add_argument("-a", "--table", dest='my_a', required=True, type=argparse.FileType('r'),
                    help="table of splice junction supports per sample with gene information "
                         "(created with pre_std.py, pre_star.py or pre_segemehl.py)")
parser.add_argument("-b", "--list", dest='my_b', required=True, type=argparse.FileType('r'),
                    help="condition to sample relation in the format: condition tab-delimiter sampleName "
                         "(you can specify a subset of the samples included in the table, "
                         "however the sampleName has to be the same as used in table) "
                         "(you can specify a label in the third column that is used (if present) "
                         "when the clustering mode is executed)")
parser.add_argument("-x", "--base_condition", dest='my_base_condition', required=True,
                    help="specify base condition (direction of change between the two conditions)")
parser.add_argument("-c", "--minsupp", dest='my_ms', type=int,
                    help="min support per splice site (at least -d samples have to show this min support)(default:10)")
parser.add_argument("-d", "--minsamples", dest='my_mt', type=int,
                    help="min amount of samples showing  in at least one of the junctions the min supp (default:3)")
parser.add_argument("-q", "--significanceThreshold", dest='my_qvalue', type=float,
                    help="significance level alpha (default 0.01)")
parser.add_argument("-z", "--foldchangeThreshold", dest='my_fc', type=float,
                    help="abundance change threshold (default 1.0)")
parser.add_argument("-e", "--cluster", action="store_true", dest='cluster',
                    help="provide -e if clustering should be performed")
parser.add_argument("-f", "--dendogram", dest='my_f', help="prefix specifying the dendogram plot")
parser.add_argument("-r", "--random", action="store_true", dest='my_r',
                    help="provide -r if random seed should be used (not deterministic mode)")


args = parser.parse_args()
if args.my_ms:
    minsupp = args.my_ms
else:
    minsupp = 10
if args.my_mt:
    min_samples = args.my_mt
else:
    min_samples = 3
if args.my_qvalue:
    myqvalue = args.my_qvalue
else:
    myqvalue = 0.01
if args.my_fc:
    myfc = args.my_fc
else:
    myfc = 1.0
perform_cluster = args.cluster
if args.my_r:
    my_seed = datetime.now().microsecond
else:
    my_seed = 102261126

program_version = "0.1.2"

sys.stderr.write("==============================================\n")
sys.stderr.write("====================DIEGO=====================\n\n")
sys.stderr.write("  _______     _____   _________     ______      ____                 \n")
sys.stderr.write(" |_      `.  |_   _| |_   ___  |  .' ___  |   .'    `.                \n")
sys.stderr.write("   |   __  \   | |     | |_  \_| /  /   \_|  |   __   |                \n")
sys.stderr.write("   |  /  \ |   | |     |  _|  _  | |   ____  |  /  \  |                 \n")
sys.stderr.write("  _|  \__/ /  _| |_   _| |___/ | \  \___]  | \  \__/  /                \n")
sys.stderr.write(" |_______.'  |_____| |_________|  `._____.'   `.____.'               \n\n")
sys.stderr.write("====================DIEGO=====================\n")
sys.stderr.write("==============================================\n")
if perform_cluster:
    if args.my_f:
        dendro_name = args.my_f
    else:
        dendro_name = "cluster_dendrogram"
    sys.stderr.write("You are running the version {} of DIEGO "
                     "(DIfferential altErnative splicinG detectiOn) in cluster mode\n".format(program_version))

else:
    sys.stderr.write("You are running the version {} of DIEGO "
                     "(DIfferential altErnative splicinG detectiOn) in normal mode\n".format(program_version))
sys.stderr.write("Parameters: minimum read support: {}, min amount of samples per condition: {}, "
                 "minimum abundance change cutoff: {}, significance threshold: {}, base condition: {}, "
                 "random seed: {} \n".format(minsupp, min_samples, myfc, myqvalue, args.my_base_condition, my_seed))

geneH = {}
scH = {}
design = []
baseC = ""
compC = ""
useClusterNames = False
sampleList = []
sampleListCluster = []
p_valueL = []
correspondingL = []


class Junction(object):
    """Subclass of Module

    Junction
    """
    left = -1
    right = -1
    significant = "none"
    p_value = -1
    q_value = -1
    ab_change = 0
    jtype = "none"
    zero_replaced = "none"
    bio_sig = -1

    def __init__(self, start, end, jtype):
        self.left = int(start)
        self.right = int(end)
        self.significant = "no"
        self.p_value = 1
        self.q_value = 1
        self.ab_change = 0
        self.jtype = jtype
        self.zero_replaced = "False"
        self.bio_sig = -1


def make_Junction(left, right, jtype):
    """initialize Junction object

    """
    myJunction = Junction(left, right, jtype)
    return myJunction


class Gene(object):
    """Subclass of Module

    Gene
    """
    ensg = ""
    name = ""
    chrom = ""
    conditions = {}
    junctionL = []
    centreH = {}
    centreH_clr = {}
    centreDist = -1

    def __init__(self, ensg, name, chrom):
        self.chrom = chrom
        self.ensg = ensg
        self.name = name
        self.conditions = {}
        self.junctionL = []
        self.centreH = {}
        self.centreH_clr = {}
        self.centreDist = -1


def make_Gene(ensg, name, chrom):
    """initialize Gene object

    """
    myGene = Gene(ensg, name, chrom)
    return myGene


class Gene_cluster(object):
    """Subclass of Module

    Genecluster
    """
    ensg = ""
    name = ""
    chrom = ""
    sampleH = {}
    sampleL = {}
    junctionL = []
    sampleDistL = []

    # the class constructor
    def __init__(self, ensg, name, chrom):
        self.chrom = chrom
        self.ensg = ensg
        self.sampleH = {}
        self.junctionL = []
        self.sampleDistL = []
        self.name = name
        self.sampleL = []


def make_Gene_cluster(ensg, name, chrom):
    """initialize Gene cluster object

    """
    myGene = Gene_cluster(ensg, name, chrom)
    return myGene


def read_sample_condition_info():
    """creates the design.

    """
    global baseC, compC
    condiH = {}
    for line in args.my_b:
        cols = line.strip().split('\t')
        if cols[1] in scH:
            sys.stderr.write("sample is multiple times present: "+str(cols[1])+" \n")
            sys.exit(1)
        else:
            scH[cols[1]] = cols[0]
            condiH[cols[0]] = 1
    if len(condiH) > 2:
        sys.stderr.write("more than two conditions specified with -b\n")
        sys.exit(1)
    else:
        for key in condiH:
            if key == args.my_base_condition:
                baseC = str(key)
                firstB = True
            else:
                compC = str(key)
        if not firstB:
            sys.stderr.write("specified base condition is not present in file provided by -b\n")
            sys.exit(1)


def read_sample_info_cluster():
    """reads cluster info name.

    """
    global useClusterNames
    for line in args.my_b:
        cols = line.strip().split('\t')
        sampleList.append(cols[1])
        scH[cols[1]] = 1
        if len(cols) > 2:
            useClusterNames = True
            sampleListCluster.append(cols[2])
        else:
            useClusterNames = False


def create_design(line):
    """creates the design.

    """
    cols = line.strip().split('\t')
    for idx, ele in enumerate(cols):
        if 1 < idx < len(cols) - 2:
            if ele in scH:
                design.append(ele)
            else:
                design.append("skip_me")


def read_table():
    """creates the design.

    """
    read_sample_condition_info()
    firstLine = True
    for line in args.my_a:
        if firstLine:
            create_design(line)
        else:
            cols = line.strip().split('\t')
            chrom = cols[0].strip().split(':')
            leftright = chrom[1].strip().split('-')
            jtype = cols[1]
            geneID = cols[-2]
            geneName = cols[-1]
            if geneID not in geneH:
                geneH[geneID] = make_Gene(geneID, geneName, chrom[0])
            geneH[geneID].junctionL.append(make_Junction(leftright[0], leftright[1], jtype))
            for idxx, ele in enumerate(cols):
                if 1 < idxx < len(cols) - 2:
                    idx = idxx - 2
                    if design[idx] != "skip_me":
                        if scH[design[idx]] not in geneH[geneID].conditions:
                            geneH[geneID].conditions[scH[design[idx]]] = {}
                        if design[idx] not in geneH[geneID].conditions[scH[design[idx]]]:
                            geneH[geneID].conditions[scH[design[idx]]][design[idx]] = []
                        geneH[geneID].conditions[scH[design[idx]]][design[idx]].append(int(ele))
        firstLine = False


def read_table_cluster():
    """creates the design.

    """
    read_sample_info_cluster()
    firstLine = True
    for line in args.my_a:
        if firstLine:
            create_design(line)
        else:
            cols = line.strip().split('\t')
            chrom = cols[0].strip().split(':')
            leftright = chrom[1].strip().split('-')
            jtype = cols[1]
            geneID = cols[-2]
            geneName = cols[-1]
            if geneID not in geneH:
                geneH[geneID] = make_Gene_cluster(geneID, geneName, chrom[0])
            junction_obj = make_Junction(leftright[0], leftright[1], jtype)
            geneH[geneID].junctionL.append(junction_obj)
            for idxx, ele in enumerate(cols):
                if 1 < idxx < len(cols) - 2:
                    idx = idxx - 2
                    if design[idx] != "skip_me":
                        if design[idx] not in geneH[geneID].sampleH:
                            geneH[geneID].sampleH[design[idx]] = []
                        geneH[geneID].sampleH[design[idx]].append(int(ele))
        firstLine = False


def zeroReplacement_multiplicative(gene):
    """replaces the zeros in count matrix.

    :type gene: Gene
    """
    mdelta = 0.5/len(gene.junctionL)
    for condi in gene.conditions:
        for sample in gene.conditions[condi]:
            comp_sum = 0
            zero_counter = 0
            for ele in gene.conditions[condi][sample]:
                comp_sum += ele
                if ele == 0:
                    zero_counter += 1
            for idx, ele in enumerate(gene.conditions[condi][sample]):
                if ele == 0:
                    gene.junctionL[idx].zero_replaced = "True"
                    gene.conditions[condi][sample][idx] = mdelta
                else:
                    gene.conditions[condi][sample][idx] = (1-zero_counter*mdelta)*(float(ele)/comp_sum)


def zeroReplacement_nbinom(gene, nbinnormalL):
    """replaces the zeros in count matrix.

    :type gene: Gene
    :type nbinnormalL: list
    """
    idx_nbinom = 0
    for condi in gene.conditions:
        sample_amount = len(gene.conditions[condi])
        count_replacementL = [0]*len(gene.junctionL)
        for sample in gene.conditions[condi]:
            comp_sum = 0
            for idx, ele in enumerate(gene.conditions[condi][sample]):
                if ele == 0:
                    idx_nbinom += 1
                    count_replacementL[idx] += 1
                    gene.conditions[condi][sample][idx] = nbinnormalL[idx_nbinom]
                comp_sum += gene.conditions[condi][sample][idx]
            for idx, ele in enumerate(gene.conditions[condi][sample]):
                gene.conditions[condi][sample][idx] = (float(ele)/comp_sum)
        for _idx in range(len(gene.junctionL)):
            if count_replacementL[_idx] >= sample_amount/2:
                gene.junctionL[_idx].zero_replaced = "True"


def zeroReplacement_nbinom_cluster(gene, nbinnormalL):
    """replaces the zeros in count matrix.

    :type gene: Gene
    :type nbinnormalL: list
    """
    sample_amount = len(gene.sampleH)
    count_replacementL = [0]*len(gene.junctionL)
    idx_nbinom = 0
    for sample in gene.sampleH:
        comp_sum = 0
        for idx, ele in enumerate(gene.sampleH[sample]):
            if ele == 0:
                idx_nbinom += 1
                count_replacementL[idx] += 1
                gene.sampleH[sample][idx] = nbinnormalL[idx_nbinom]
            comp_sum += gene.sampleH[sample][idx]
        for idx, ele in enumerate(gene.sampleH[sample]):
            gene.sampleH[sample][idx] = (float(ele)/comp_sum)
    for _idx in range(len(gene.junctionL)):
        if count_replacementL[_idx] >= sample_amount/2:
            gene.junctionL[_idx].zero_replaced = "True"


def reduceMatrix(gene, nbinnormalL):
    """reduces the count matrix.

    :type gene: Gene
    :type nbinnormalL: list
    :rtype reference_file: Bool
    """
    minjunctions = 2
    maxjunctions = 60
    keep_colH = {}
    keep_colHpre = {}
    for condi in gene.conditions:
        toRemoveC = []
        for sample in gene.conditions[condi]:
            keep_composition = False
            for idx, ele in enumerate(gene.conditions[condi][sample]):
                if ele >= minsupp:
                    keep_composition = True
                    if idx not in keep_colHpre:
                        keep_colHpre[idx] = 0
                    keep_colHpre[idx] += 1
            if not keep_composition:
                toRemoveC.append(sample)
        for sample in toRemoveC:
            del gene.conditions[condi][sample]
    for key, value in keep_colHpre.items():
        if value >= min_samples:
            keep_colH[key] = 1
    myIndices = sorted(keep_colH.keys())
    for condi in gene.conditions:
        for sample in gene.conditions[condi]:
            gene.conditions[condi][sample] = [gene.conditions[condi][sample][i] for i in myIndices]
    gene.junctionL = [gene.junctionL[i] for i in myIndices]
    if len(gene.junctionL) < minjunctions or len(gene.junctionL) > maxjunctions:
        return True
    else:
        for condi in gene.conditions:
            if len(gene.conditions[condi]) < min_samples:
                return True
    zeroReplacement_nbinom(gene, nbinnormalL)
    return False


def reduceMatrix_cluster(gene, nbinnormalL):
    """reduces the count matrix.

    :type gene: Gene
    :type nbinnormalL: list
    :rtype reference_file: Bool
    """
    minjunctions = 2
    maxjunctions = 60
    keep_colH = {}
    keep_colHpre = {}
    toRemoveC = []
    for sample in gene.sampleH:
        keep_composition = False
        for idx, ele in enumerate(gene.sampleH[sample]):
            if ele >= minsupp:
                keep_composition = True
                if idx not in keep_colHpre:
                    keep_colHpre[idx] = 0
                keep_colHpre[idx] += 1
        if not keep_composition:
            toRemoveC.append(sample)
    for sample in toRemoveC:
        del gene.sampleH[sample]
    for key, value in keep_colHpre.items():
        if value >= min_samples:
            keep_colH[key] = 1
    myIndices = sorted(keep_colH.keys())
    for sample in gene.sampleH:
        gene.sampleH[sample] = [gene.sampleH[sample][i] for i in myIndices]
    gene.junctionL = [gene.junctionL[i] for i in myIndices]
    if len(gene.junctionL) < minjunctions or len(gene.junctionL) > maxjunctions:
        return True
    else:
        if len(gene.sampleH) < min_samples:
            return True
    zeroReplacement_nbinom_cluster(gene, nbinnormalL)
    return False


def dist_eukledian(firstL, secondL):
    """calc eukledian
    :type firstL: list
    :type secondL: list
    :rtype : float
    """
    diste = 0
    for i in range(len(firstL)):
        diste += math.pow(firstL[i] - secondL[i], 2)
    return math.pow(diste, 0.5)


def build_centre_clr_transformed_cluster(gene):
    """build cluster

    :type gene: Gene
    :rtype center_list: list
    """
    center_list = []
    for i in range(len(gene.junctionL)):
        column_values = []
        for sample in gene.sampleH:
            column_values.append(gene.sampleH[sample][i])
        center_list.append(arithmean(column_values))
    return center_list


def build_centre_clr_transformed(gene, condi):
    """Info
    :type gene: Gene
    :type condi: str
    """
    center_list = []
    for i in range(len(gene.junctionL)):
        column_values = []
        for sample in gene.conditions[condi]:
            column_values.append(gene.conditions[condi][sample][i])
        center_list.append(arithmean(column_values))
    return center_list


def geomean(my_list):
    """calc geomean
    :type my_list: list
    :rtype : float
    """
    gs = 1.0
    for ele in my_list:
        gs = gs * ele
    if gs <= 0.0:
        sys.stderr.write(" list product is zero or negative! (in geomean()) \n")
        sys.stderr.write(str(my_list)+"\n")
        sys.exit(1)
    elif len(my_list) == 0:
        sys.stderr.write(" list is empty (in geomean()) \n")
        sys.stderr.write(str(my_list)+"\n")
    else:
        return math.pow(gs, 1.0/len(my_list))


def arithmean(my_list):
    """calc geomean
    :type my_list: list
    :rtype : float
    """
    if len(my_list) > 0:
        return float(sum(my_list))/len(my_list)
    else:
        sys.stderr.write(" ERROR empty List to arithmetic mean")
        sys.exit(1)


def transform(gene):
    """transform
    :type gene: Gene
    """
    for condi in gene.conditions:
        for sample in gene.conditions[condi]:
            geom_mean = geomean(gene.conditions[condi][sample])
            for idx, ele in enumerate(gene.conditions[condi][sample]):
                gene.conditions[condi][sample][idx] = (math.log((ele/geom_mean)))


def transform_cluster(gene):
    """transform
    :type gene: Gene
    """
    for sample in gene.sampleH:
        geom_mean = geomean(gene.sampleH[sample])
        for idx, ele in enumerate(gene.sampleH[sample]):
            gene.sampleH[sample][idx] = (math.log((ele/geom_mean)))


def dist_simplex(firstL, secondL):
    """calc dist
    :type firstL: list
    :type secondL: list
    :rtype : float
    """
    geom_first = geomean(firstL)
    geom_second = geomean(secondL)
    diste = 0
    for i in range(len(firstL)):
        diste += math.pow((math.log((firstL[i]/geom_first))-math.log((secondL[i]/geom_second))), 2)
    return pow(diste, 0.5)


def build_centre(gene, condi):
    """build centre
    :type gene: Gene
    :type condi: str
    :rtype : list
    """
    center_list = []
    for i in range(len(gene.junctionL)):
        column_values = []
        for sample in gene.conditions[condi]:
            column_values.append(gene.conditions[condi][sample][i])
        if len(column_values) == 0:
            sys.stderr.write(" ERROR empty list to geom mean \n")
            sys.exit(1)
        center_list.append(geomean(column_values))
    geom_sum = sum(center_list)
    center_list_result = [x / geom_sum for x in center_list]
    return center_list_result


def dist_centre_2_centre(gene):
    """build centre
    :type gene: Gene
    """
    for condis in gene.conditions:
        gene.centreH[condis] = build_centre(gene, condis)
    myDist = dist_simplex(gene.centreH[baseC], gene.centreH[compC])
    gene.centreDist = myDist


def dist_all_sample_2_centre_cluster(gene):
    """build centre
    :type gene: Gene
    :rtype : float
    """
    distance = 0
    centre1 = build_centre_clr_transformed_cluster(gene)
    for sample in gene.sampleH:
        if sample in gene.sampleH:
            distance += dist_eukledian(centre1, gene.sampleH[sample])
    return distance/len(gene.sampleH)


def printDists(genes):
    """print distances
    :type gene: Gene
    """
    sys.stdout.write(genes.ensg+"\t"+str(genes.centreDist)+"\n")
    sys.stdout.write("centre_"+str(baseC)+":\t"+str(genes.centreH[baseC])+"\n")
    sys.stdout.write("centre_"+str(compC)+":\t"+str(genes.centreH[compC])+"\n")


def dist_centre_2_centre_clr_transformed(gene):
    """calc
    :type gene: Gene
    """
    for condis in gene.conditions:
        gene.centreH_clr[condis] = build_centre_clr_transformed(gene, condis)
    myDist = dist_eukledian(gene.centreH_clr[baseC], gene.centreH_clr[compC])
    gene.centreDist = myDist


def benjaminihochberg_correction(pvalues):
    """calc q values
    :type pvalues: list
    :rtype new_pvalues: list
    """
    n = float(len(pvalues))
    new_pvalues = list(range(int(n)))
    values = [(pvalue, i) for i, pvalue in enumerate(pvalues)]
    values.sort(reverse=True)
    new_values = []
    for i, vals in enumerate(values):
        rank = n - i
        pvalue, index = vals
        new_values.append((n/rank) * pvalue)
    for i in range(0, int(n)-1):
        if new_values[i] < new_values[i+1]:
            new_values[i+1] = new_values[i]
    for i, vals in enumerate(values):
        pvalue, index = vals
        new_pvalues[index] = new_values[i]
    return new_pvalues


def test_juctions(gene):
    """calc
    :type gene: Gene
    """
    for indx, junction in enumerate(gene.junctionL):
        junction.ab_change = gene.centreH_clr[baseC][indx]-gene.centreH_clr[compC][indx]
        if gene.centreH_clr[baseC][indx] >= gene.centreH_clr[compC][indx]:
            junction.bio_sig = gene.centreH_clr[baseC][indx]
        else:
            junction.bio_sig = gene.centreH_clr[compC][indx]
        if abs(junction.ab_change) >= myfc:
            condition_base = []
            condition_compare = []
            for sample in gene.conditions[baseC]:
                condition_base.append(gene.conditions[baseC][sample][indx])
            for sample in gene.conditions[compC]:
                condition_compare.append(gene.conditions[compC][sample][indx])
            p_value = ranksums(condition_base, condition_compare)[1]
            p_valueL.append(p_value)
            gene.junctionL[indx].p_value = p_value
            myTuple = (gene.ensg, indx)
            correspondingL.append(myTuple)


def printJunctions(genes):
    """calc
    :type gene: Gene
    """
    sig_juncs = []
    for junction in genes.junctionL:
        if junction.significant == "yes":
            sig_juncs.append(junction)
    for junction in genes.junctionL:
        sys.stdout.write(str(genes.chrom)+":"+str(junction.left)+"-"+str(junction.right)+"\t")
        sys.stdout.write(junction.jtype+"\t")
        sys.stdout.write(str(junction.ab_change)+"\t")
        sys.stdout.write(str(junction.p_value)+"\t"+str(junction.q_value)+"\t")
        sys.stdout.write(genes.ensg+"\t"+genes.name+"\t"+str(len(genes.junctionL))+"\t"+str(len(sig_juncs))+"\t"+str(genes.centreDist)+"\t")
        sys.stdout.write(junction.significant+"\t"+junction.zero_replaced+"\n")


def printMatrix(gene):
    """calc
    :type gene: Gene
    """
    sys.stderr.write(gene.ensg+"\n")
    sys.stderr.write("gene _junction_laenge:: "+str(str(len(gene.junctionL)))+"\n")
    sys.stderr.write("gene_condition_laenge:: "+str(str(len(gene.conditions)))+"\n")
    for condi in gene.conditions:
        sys.stderr.write("condi: "+str(condi)+"\n")
        for sample in gene.conditions[condi]:
            sys.stderr.write(sample+"\t"+str(gene.conditions[condi][sample])+"\n")


def calculate_pairwise_dist_matrix():
    """calc  pairwise_dist_matrix
    """
    amount_of_genes = len(geneH)
    samplePairL = list(itertools.combinations(sampleList, 2))
    cond_dist_matrix = np.zeros((len(samplePairL),), dtype=np.float)
    for idx, samplePair in enumerate(samplePairL):
        substractCounter = 0
        samplePairDist = 0
        for gene in geneH.values():
            if samplePair[0] in gene.sampleH and samplePair[1] in gene.sampleH:
                samplePairDist += dist_eukledian(gene.sampleH[samplePair[0]], gene.sampleH[samplePair[1]])
            else:
                substractCounter += 1
        cond_dist_matrix[idx] = samplePairDist/(amount_of_genes-substractCounter)
    linkageResult = linkage(cond_dist_matrix, method='average')
    if useClusterNames:
        #dendrogram(linkageResult, color_threshold=3.46, leaf_font_size=4,labels = sampleListCluster)
        dendrogram(linkageResult, leaf_font_size=4, labels=sampleListCluster)
    else:
        dendrogram(linkageResult, leaf_font_size=4, labels=sampleList)
    plt.savefig(dendro_name+".pdf")


def find_interesting_genes():
    """find_interesting_genes

    """
    genNameVarL = []
    for gene in geneH.values():
        genNameVarL.append((gene.ensg, dist_all_sample_2_centre_cluster(gene)))
    genNameVarL.sort(key=operator.itemgetter(1))
    for ele in genNameVarL[:int(len(genNameVarL) * 0.75)]:
        del geneH[ele[0]]


def perform_clustering():
    """perform clustering

    """
    #import matplotlib
    #matplotlib.use('Agg')
    #import matplotlib.pyplot as plt
    
    nbinnormalL = nbinom.rvs(1, 0.4, loc=1, size=1000000, random_state=my_seed)
    toRemoveG = []
    read_table_cluster()
    sys.stderr.write("Number of genes before matrix reduction: "+str(len(geneH))+"\n")
    for gene in geneH.values():
        if reduceMatrix_cluster(gene, nbinnormalL):
            toRemoveG.append(gene.ensg)
    for ele in toRemoveG:
        del geneH[ele]
    sys.stderr.write("Number of genes remaining: "+str(len(geneH))+"\n")
    if len(geneH) < 1:
        sys.exit(1)
    for gene in geneH.values():
        transform_cluster(gene)
    find_interesting_genes()
    calculate_pairwise_dist_matrix()


def perform_DAS_detection():
    """run algo

    """
    sig_counter = 0
    nbinnormalL = nbinom.rvs(1, 0.4, loc=1, size=1000000, random_state=my_seed)
    toRemoveG = []
    read_table()
    sys.stderr.write("Number of genes before matrix reduction: "+str(len(geneH))+"\n")
    for gene in geneH.values():
        if reduceMatrix(gene, nbinnormalL):
            toRemoveG.append(gene.ensg)
    for ele in toRemoveG:
        del geneH[ele]
    sys.stderr.write("Number of genes remaining: "+str(len(geneH))+"\n")
    if len(geneH) < 1:
        sys.exit(1)
    for gene in geneH.values():
        transform(gene)
        dist_centre_2_centre_clr_transformed(gene)
        test_juctions(gene)
    q_valueL = benjaminihochberg_correction(p_valueL)
    for idx, ele in enumerate(correspondingL):
        geneH[ele[0]].junctionL[ele[1]].q_value = q_valueL[idx]
        if q_valueL[idx] < myqvalue:
            geneH[ele[0]].junctionL[ele[1]].significant = "yes"
            sig_counter += 1
    sys.stdout.write("junction\tjunction_type\tabundance_change\t")
    sys.stdout.write("p_val\tq_val\tgeneID\tgeneName\tjunctions\tsig_junctions\tcentre_dist\tsignificant\tzero_replacement\n")
    for gene in geneH.values():
        printJunctions(gene)
    sys.stderr.write('Number of significant differential alternative splicing events: '+str(sig_counter)+'\n')


def run_tool():
    """run tool

    """
    if perform_cluster:
        perform_clustering()
    else:
        perform_DAS_detection()
    sys.stderr.write('Have fun with your results. \n')
    sys.stdout.flush()
    sys.stdout.close()


run_tool()



#!/usr/bin/env python3

import sys
import argparse
import os
import inspect
import gzip
import glob
import re
import logging
import traceback as tb

cmd_subfolder = os.path.join(
    os.path.dirname(
        os.path.realpath(os.path.abspath(inspect.getfile(inspect.currentframe())))
    ),
    "../../lib",
)
if cmd_subfolder not in sys.path:
    sys.path.insert(0, cmd_subfolder)

from Logger import *

try:
    scriptname = os.path.basename(inspect.stack()[-1].filename).replace(".py", "")
    log = logging.getLogger(scriptname)

    lvl = log.level if log.level else "INFO"
    for handler in log.handlers[:]:
        handler.close()
        log.removeHandler(handler)

    try:
        handler = logging.FileHandler("LOGS/MONSDA.log", mode="a")
    except:
        handler = logging.FileHandler("log", mode="a")

    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    log.setLevel(lvl)

except Exception:
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = tb.TracebackException(
        exc_type,
        exc_value,
        exc_tb,
    )
    print("".join(tbe.format()), file=sys.stderr)


def parseargs():
    parser = argparse.ArgumentParser(description="Creates contrast list for DIEGO")
    parser.add_argument("-a", "--annofile", type=str, default="", help="")
    parser.add_argument("-b", "--combi", type=str, default="", help="")
    parser.add_argument("-c", "--comparisons", type=str, default="", help="")
    parser.add_argument("-o", "--outdir", type=str, default="", help="")
    parser.add_argument("--loglevel", default="INFO", help="Log verbosity")
    parser.add_argument("--nextflow", action="store_true", help="Run in nextflow mode")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


def create_tables(annofile, combi, comparisons, outdir):
    logid = scriptname + ".create_tables: "

    if combi == "none":
        combi = ""

    comps = comparisons.split(",")
    sample_dict = {}
    with open(annofile, "r") as gf:
        for line in gf:
            l = line.replace("\n", "")
            if l == "":
                continue
            sample_dict[l.split("\t")[0]] = l.split("\t")[1]

    for c in comps:
        log.debug(f"{logid}COMPS: {c}")
        contrast_name = c.split(":")[0]
        contrast_group1 = [i for i in c.split(":")[1].split("-vs-")[0].split("+")]
        contrast_group2 = [i for i in c.split(":")[1].split("-vs-")[1].split("+")]

        outstring = ""
        for sample in sample_dict:
            if sample_dict[sample] in contrast_group1:
                outstring += f"{contrast_name}_1\t{sample}\n"
            if sample_dict[sample] in contrast_group2:
                outstring += f"{contrast_name}_2\t{sample}\n"
        with open(f"{outdir}/{combi}_{contrast_name}_contrast.txt", "w") as outfile:
            outfile.write(outstring)


####################
####    MAIN    ####
####################

if __name__ == "__main__":
    logid = scriptname + ".main: "
    try:
        args = parseargs()

        try:
            if not args.nextflow:
                makelogdir("LOGS")
                log = setup_logger(
                    name=scriptname,
                    log_file="LOGS/" + scriptname + ".log",
                    logformat="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
                    datefmt="%m-%d %H:%M",
                    level=args.loglevel,
                )
            else:
                log = setup_logger(
                    name=scriptname,
                    log_file="log",
                    logformat="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
                    datefmt="%m-%d %H:%M",
                    level=args.loglevel,
                )
            log.addHandler(logging.StreamHandler(sys.stderr))  # streamlog
        except:
            log = logging.getLogger(os.path.basename(inspect.stack()[-1].filename))

        log.info("RUNNING!")
        create_tables(args.annofile, args.combi, args.comparisons, args.outdir)

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        log.error(logid + "".join(tbe.format()))


# diego_contrast_files.py ends here



#!/usr/bin/env python3
import re
import subprocess as sp
import shlex
import sys
import time
import logging

logger = logging.getLogger("__name__")

STATUS_ATTEMPTS = 20

jobid = sys.argv[1]

  # noqa: E999,E225
cluster = "--cluster=cluster"
  # noqa: E225

for i in range(STATUS_ATTEMPTS):
    try:
        sacct_res = sp.check_output(shlex.split(f"sacct {cluster} -P -b -j {jobid} -n"))
        res = {
            x.split("|")[0]: x.split("|")[1]
            for x in sacct_res.decode().strip().split("\n")
        }
        break
    except sp.CalledProcessError as e:
        logger.error("sacct process error")
        logger.error(e)
    except IndexError as e:
        pass
    # Try getting job with scontrol instead in case sacct is misconfigured
    try:
        sctrl_res = sp.check_output(
            shlex.split(f"scontrol {cluster} -o show job {jobid}")
        )
        m = re.search(r"JobState=(\w+)", sctrl_res.decode())
        res = {jobid: m.group(1)}
        break
    except sp.CalledProcessError as e:
        logger.error("scontrol process error")
        logger.error(e)
        if i >= STATUS_ATTEMPTS - 1:
            print("failed")
            exit(0)
        else:
            time.sleep(1)

status = res[jobid]

if status == "BOOT_FAIL":
    print("failed")
elif status == "OUT_OF_MEMORY":
    print("failed")
elif status.startswith("CANCELLED"):
    print("failed")
elif status == "COMPLETED":
    print("success")
elif status == "DEADLINE":
    print("failed")
elif status == "FAILED":
    print("failed")
elif status == "NODE_FAIL":
    print("failed")
elif status == "PREEMPTED":
    print("failed")
elif status == "TIMEOUT":
    print("failed")
# Unclear whether SUSPENDED should be treated as running or failed
elif status == "SUSPENDED":
    print("failed")
else:
    print("running")



#!/usr/bin/env python3
# test_functions.py ---
#
# Filename: test_functions.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Wed Feb 19 12:31:13 2020 (+0100)
# Version:
# Package-Requires: ()
# Last-Updated: Tue Feb 25 21:31:50 2020 (+0100)
#           By: Joerg Fallmann
#     Update #: 79
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:
import glob
import os
import sys
import inspect
import json
import shutil
from collections import defaultdict
import traceback as tb
from snakemake.common.configfile import load_configfile
from snakemake.utils import validate
import argparse
import subprocess
import re

from MONSDA.Collection import *
from MONSDA.Logger import *

print(sys.argv)

if len(sys.argv) < 2:
    level = 'DEBUG'
else:
    level = sys.argv[1]

log = setup_logger(name='', log_file='stderr', logformat='%(asctime)s %(name)-12s %(levelname)-8s %(message)s', datefmt='%m-%d %H:%M', level=level)
logid = 'TESTER: '

config = load_configfile('/home/fall/Work/Tests/SnakemakeTest/SubSnakes/ID_unpaired_std_subconfig.json') if len(sys.argv) < 3 else load_configfile(sys.argv[2])
REFERENCE=config['REFERENCE']

file=r'ID/unpaired/std/GSM461177_untreat_paired_subset_r1'
dir = 'Dm6'

SAMPLES=list()
SAMPLES.append(file)

print(logid+'SAMPLES: '+str(SAMPLES))

print(logid+'GENOME: '+genome(file, config))

print(logid+'TESTOPTIONS: '+str(tool_params(file, None, config, 'MAPPING')))

print(logid+'OPTIONS: '+ tool_params(SAMPLES[0], None, config, 'MAPPING')['OPTIONS']['INDEX']

print(logid+'ANNO: '+str.join(os.sep,[config["REFERENCE"],os.path.dirname(genomepath(SAMPLES[0], config)),tool_params(SAMPLES[0], None, config, 'MAPPING')['ANNOTATION']]))

MAPPERBIN, MAPPERENV = env_bin_from_config(SAMPLES,config,'MAPPING')
print(logid+'ENV/BIN: '+str([MAPPERENV,MAPPERBIN]))
print(logid+'GENPATH: '+"{ref}/{dir}/{map}/{extension}/".format(ref=REFERENCE, dir=dir, map=MAPPERENV, extension=check_tool_params(SAMPLES[0], None,config, 'MAPPING','MAP')))

print(logid+'ref: '+"{ref}/{dir}/{gen}{name}".format(ref=REFERENCE, dir = source_from_sample(SAMPLES[0],config), gen =genome(file, config), name=namefromfile(file, config)))

print(logid+'idx: '+"{ref}/{dir}/{gen}{name}/{map}_{extension}".format(ref=REFERENCE, dir=source_from_sample(file,config), gen=genome(file, config), name=namefromfile(file, config), map=MAPPERENV, extension=check_tool_params(file, None,config, 'MAPPING','EXTENSION')))

test = {'a':{'a1':{'a11':{'a111'},'a12':{'a121'}},'a2':{'a22'}},'b':{'b1':{'b2':{'b3'}}}}

print(logid+'Keysettest: '+str(test))
#print(keys_from_dict(test))
print(logid+'Keyset: '+str(keysets_from_dict(test)))

print(logid+'DEBINENV: '+str(env_bin_from_config(SAMPLES,config,'MAPPING')))



#print(logid+)

#
# test_functions.py ends here



# CountEnds.py ---
#
# Filename: CollectBamStat.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Tue May 15 12:44:06 2018 (+0200)
# Version:
# Package-Requires: ()
# Last-Updated: Wed Jan 30 15:53:25 2019 (+0100)
#           By: Joerg Fallmann
#     Update #: 923
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:

### IMPORTS
import argparse
import pprint  # pp = pprint.PrettyPrinter(indent=4) #pp.pprint(stuff)
from io import StringIO
import os
import gzip
import re
import sys
import pysam
from pyfaidx import Fasta
import logging

# import importlib
from multiprocessing import Pool, Manager
import traceback as tb

# import pickle
# Container
import collections
import inspect

try:
    scriptname = os.path.basename(inspect.stack()[-1].filename).replace(".py", "")
    log = logging.getLogger(scriptname)

    lvl = log.level if log.level else "INFO"
    for handler in log.handlers[:]:
        handler.close()
        log.removeHandler(handler)

    handler = logging.FileHandler("LOGS/MONSDA.log", mode="a")
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    log.setLevel(lvl)

except Exception:
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = tb.TracebackException(
        exc_type,
        exc_value,
        exc_tb,
    )
    print("".join(tbe.format()), file=sys.stderr)


### MAIN
def parseargs():
    parser = argparse.ArgumentParser(
        description="Read BAM file read by read and collect mapping statistics"
    )
    parser.add_argument(
        "-f", "--fasta", type=str, default=None, help="Fasta with cluster info"
    )
    parser.add_argument(
        "-b",
        "--bams",
        type=str,
        help="Mapped reads BAM(s) comma separated, need to besorted and indexed",
    )
    parser.add_argument(
        "-l",
        "--lookup",
        type=str,
        default=None,
        help="Search for given ID in chrom tag of reads, e.g. Cluster, chr1, etc.",
    )
    parser.add_argument(
        "-t", "--offset", type=int, default=0, help="Offset for search around read end"
    )
    parser.add_argument(
        "-o", "--outdir", type=str, default=".", help="Output directory"
    )
    parser.add_argument(
        "-z",
        "--procs",
        type=int,
        default=1,
        help="Number of parallel processed to run this job with",
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        type=int,
        default=0,
        choices=[0, 1],
        help="increase output verbosity",
    )

    return parser.parse_args()


def collectends(bams, outdir, procs, verbosity, offset, fasta=None, lookup=None):
    try:
        if outdir:
            print("Checking or creating outdir " + str(outdir))
            if not os.path.isabs(outdir):
                outdir = os.path.abspath(outdir)
            if not os.path.exists(outdir):
                os.makedirs(outdir)
        else:
            outdir = os.path.abspath(os.getcwd())

        ###Multithreading
        num_processes = procs or 1
        pool = Pool(processes=num_processes, maxtasksperchild=1)
        collector = collections.OrderedDict()
        translater = None

        if fasta:
            translater = cluster2trna(fasta)[
                "cluster"
            ]  # Make sure to use the right fasta for the bams

        for bam in bams.split(","):
            fn = os.path.basename(bam)
            out = os.path.join(outdir, fn + ".ends.gz")
            collector[fn] = collections.OrderedDict()
            collector[fn]["out"] = out
            collector[fn]["res"] = collections.OrderedDict()

            if os.path.isfile(out):
                print("File " + out + " exists, will be deleted")
                os.remove(out)

            write_header(out, translater)

            if os.path.exists(bam) and os.path.getsize(bam) > 0:
                samfile = parse_bam(bam)
                try:
                    ####Check if index files available
                    samfile.check_index()
                except:
                    print("No index for file: " + bam + ". Please create!")
                    sys.exit()

                header = None
                try:
                    if "SQ" in samfile.header:
                        header = samfile.header["SQ"]
                except:
                    print(
                        "Could not read header of file "
                        + bam
                        + ". Multithreading only per file not per chromosome."
                    )

                if header:
                    for i, chrom in enumerate(h["SN"] for h in header):
                        if lookup:
                            if lookup not in chrom:
                                continue
                        res = pool.apply_async(
                            collect, args=(bam, offset, chrom), error_callback=eprint
                        )
                        rescheck = res.get()
                        if chrom in rescheck:
                            collector[fn]["res"][chrom] = rescheck[chrom]

                else:
                    res = pool.apply_async(
                        collect, args=(bam, offset), error_callback=eprint
                    )
                    rescheck = res.get()
                    if len(rescheck) > 0:
                        collector[fn]["res"] = rescheck
            else:
                write_empty(out)
        pool.close()
        pool.join()

        for bam in collector:
            write_stats(collector[bam], translater)

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def collect(bam, offset, chrom=None):
    # READS HAVE TO MAP TO END OF CLUSTER
    # TODO: CHECK IF NUMBERS ARE CORRECT!
    try:
        stats = collections.OrderedDict()

        samfile = parse_bam(bam)
        refs = dict(zip(samfile.references, samfile.lengths))

        if chrom and chrom is not "None":
            print("Parsing Chromosome parallel " + str(chrom))
            for read in samfile.fetch(chrom):
                #                if not read.is_unmapped:                        # This currently doesn't work with segemehl!
                seq = read.query_alignment_sequence.upper()  # read without softclip
                rawseq = read.query_sequence.upper()  # read with softclip
                refseq = (
                    read.get_reference_sequence().upper()
                )  # reference part of alignment

                if read.reference_end in range(
                    refs[chrom] - offset, refs[chrom] - offset + 6 + 1
                ):  # Should make sure we are at the read end but also allow softclips
                    tags = dict(read.get_tags())
                    alignmentend = str(seq[-6:])
                    rawend = str(rawseq[-6:])
                    refend = str(refseq[-6:])

                    if not chrom in stats:
                        stats[chrom] = collections.OrderedDict()
                        stats[chrom]["seq"] = collections.OrderedDict()
                        stats[chrom]["raw"] = collections.OrderedDict()
                        stats[chrom]["ref"] = collections.OrderedDict()

                    if (
                        alignmentend in stats[chrom]["seq"]
                    ):  # Collect ends and count them
                        stats[chrom]["seq"][alignmentend] += 1 / tags["NH"]
                    else:
                        stats[chrom]["seq"][alignmentend] = 1 / tags["NH"]
                    if rawend in stats[chrom]["raw"]:  # Collect ends and count them
                        stats[chrom]["raw"][rawend] += 1 / tags["NH"]
                    else:
                        stats[chrom]["raw"][rawend] = 1 / tags["NH"]
                    if refend in stats[chrom]["ref"]:  # Collect ends and count them
                        stats[chrom]["ref"][refend] += 1 / tags["NH"]
                    else:
                        stats[chrom]["ref"][refend] = 1 / tags["NH"]

        else:
            for read in samfile.fetch():
                #                if not read.is_unmapped:  # This currently doesn's work with segemehl!
                chrom = read.reference_name
                seq = read.query_alignment_sequence.upper()
                rawseq = read.query_sequence.upper()
                refseq = (
                    read.get_reference_sequence().upper()
                )  # reference part of alignment

                if read.reference_end in range(
                    refs[chrom] - offset, refs[chrom] - offset + 6 + 1
                ):  # Should make sure we are at the read end but also allow softclips
                    #                    if refs[chrom]-offset == read.reference_end:
                    tags = dict(read.get_tags())
                    alignmentend = str(seq[-6:])
                    rawend = str(rawseq[-6:])
                    refend = str(refseq[-6:])

                    if not chrom in stats:
                        stats[chrom] = collections.OrderedDict()
                        stats[chrom]["seq"] = collections.OrderedDict()
                        stats[chrom]["raw"] = collections.OrderedDict()
                        stats[chrom]["ref"] = collections.OrderedDict()

                    if (
                        alignmentend in stats[chrom]["seq"]
                    ):  # Collect ends and count them
                        stats[chrom]["seq"][alignmentend] += 1 / tags["NH"]
                    else:
                        stats[chrom]["seq"][alignmentend] = 1 / tags["NH"]
                    if rawend in stats[chrom]["raw"]:  # Collect ends and count them
                        stats[chrom]["raw"][rawend] += 1 / tags["NH"]
                    else:
                        stats[chrom]["raw"][rawend] = 1 / tags["NH"]
                    if refend in stats[chrom]["ref"]:  # Collect ends and count them
                        stats[chrom]["ref"][refend] += 1 / tags["NH"]
                    else:
                        stats[chrom]["ref"][refend] = 1 / tags["NH"]

        close_bam(samfile)

        return stats

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def fillre(col):
    try:
        rc = col.get()
        return rc

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def collect_collector(bam, chrom=None):

    try:
        rc = readcollector(os.path.basename(bam))
        samfile = parse_bam(bam)
        if chrom and chrom is not "None":
            printlog("Parsing Chromosome parallel" + str(chrom))
            for read in samfile.fetch(chrom):
                if not read.is_unmapped:
                    seq = read.query_alignment_sequence
                    tags = dict(read.get_tags())
                    alignmentend = str(seq[-6:])
                    if not rc.checkchrom(chrom):
                        rc.add(chrom)
                    if rc.checkend(
                        [chrom, alignmentend]
                    ):  # Collect ends and count them
                        rc.addval([chrom, alignmentend], 1 / tags["NH"])
                    else:
                        rc.add([chrom, alignmentend], 1 / tags["NH"])
        #                        rc.addend(chrom, alignmentend, 1/tags['NH'])

        else:
            for read in samfile.fetch():
                if not read.is_unmapped:
                    chrom = read.reference_name
                    seq = read.query_alignment_sequence
                    tags = dict(read.get_tags())
                    alignmentend = seq[-6:]
                    if not rc.checkchrom(chrom):
                        rc.add(chrom)
                    if rc.checkend(
                        [chrom, alignmentend]
                    ):  # Collect ends and count them
                        rc.addval([chrom, alignmentend], 1 / tags["NH"])
                    else:
                        rc.add([chrom, alignmentend], 1 / tags["NH"])
        #           close_bam(samfile)
        return rc

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def parse_bam(bam):
    try:
        if ".bam" in bam:
            return pysam.AlignmentFile(bam, "rb")
        elif ".sam" in bam:
            if ".gz" in bam[-4:]:
                return pysam.AlignmentFile(gzip.open(bam, "rt"), "r")
            else:
                return pysam.AlignmentFile(bam, "r")
    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def parse_cigar(cigarstring, length, start):

    pattern = re.compile("([0-9]*)([DMINSHP=XB])")
    stop = start

    for n, c in pattern.findall(cigar):
        if n:
            n = int(n)
        else:
            n = 1

        if c == "M" or c == "X" or c == "=":
            stop += n
        if c is "D":
            stop += n
            length += n
        if c is "N":
            stop += n

        if c in ["I"]:
            char += n
        if c in ["H", "P", "B", "S"]:  # No idea what PB mean
            continue

    return


def read_head(bam):
    try:
        if ".bam" in bam:
            return pysam.AlignmentFile(bam, "rb").header
        elif ".sam" in bam:
            return pysam.AlignmentFile(bam, "r").header
    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def write_bam(bam, template):
    try:
        return pysam.AlignmentFile(bam, "wh", template=template)
    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def close_bam(samfile):
    samfile.close()


def check_idx(file):
    try:
        if (
            (os.path.isfile(file + ".idx"))
            or (os.path.isfile(str.join(".", split(".", file)[0:-1], "fai")))
            or (os.path.isfile(str.join(".", split(".", file)[0:-1], "faidx")))
        ):
            return True
        else:
            return False
    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def get_stats(chrom, alignmentend, statistics):
    try:
        #   with open('bla','a') as h:
        #       print('Fetching Stats', file=h)

        if chrom not in statistics["ends"]:
            statistics["ends"][chrom] = collections.OrderedDict()

        for n, c in pattern.findall(cigar):
            if n:
                n = int(n)
            else:
                n = 1

            if c == "M" or c == "X" or c == "=":
                for i in range(n):
                    if pos not in statistics["reads"][chrom]:
                        statistics["reads"][chrom][pos] = collections.OrderedDict()
                    if seq[char] not in statistics["reads"][chrom][pos]:
                        statistics["reads"][chrom][pos][seq[char]] = 0
                    statistics["reads"][chrom][pos][seq[char]] += 1
                    char += 1
                    pos += 1
            if c in ["D", "N"]:
                pos += n
            if c in ["I"]:
                char += n
            if c in ["H", "P", "B", "S"]:  # No idea what PB mean
                continue

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def get_ref(fa, statistics):
    try:
        faseq = Fasta(fa)

        for chrom in statistics["reads"]:
            if chrom not in statistics["ref"]:
                statistics["ref"][chrom] = collections.OrderedDict()
            for pos in statistics["reads"][chrom]:
                if pos not in statistics["ref"][chrom]:
                    statistics["ref"][chrom][pos] = collections.OrderedDict()
                nuc = faseq[chrom][pos - 1].seq.upper()
                statistics["ref"][chrom][pos] = nuc
    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def write_header(out, fastadict=None):
    try:
        if fastadict:
            head = str.join(
                "\t",
                [
                    "Chromosome",
                    "Seq",
                    "End",
                    "Count",
                    "ChromRelRefFreq",
                    "TotalRelRefFreq",
                    "tRNAs",
                ],
            )
        else:
            head = str.join(
                "\t",
                [
                    "Chromosome",
                    "Seq",
                    "End",
                    "Count",
                    "ChromRelRefFreq",
                    "TotalRelRefFreq",
                ],
            )

        with gzip.open(out, "wb") as o:
            o.write(bytes(head + "\n", encoding="UTF-8"))

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def write_empty(out):
    try:
        with gzip.open(out, "ab") as o:
            o.write(bytes("0\n", encoding="UTF-8"))

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def write_stats(stat, fastadict=None):
    try:
        coverage = ""
        allends = collections.OrderedDict()
        totalcount = {}
        readnum = {}
        if stat is not None:
            out = stat.pop("out", None)  # get key and remove from dict
            statistics = stat.pop("res", None)  # get key and remove from dict

            for chrom in statistics:
                if chrom not in allends:
                    allends[chrom] = collections.OrderedDict()
                    allends[chrom]["seq"] = collections.OrderedDict()
                    allends[chrom]["raw"] = collections.OrderedDict()
                    allends[chrom]["ref"] = collections.OrderedDict()
                if chrom not in totalcount:
                    totalcount[chrom] = collections.OrderedDict()
                    totalcount[chrom]["seq"] = 0
                    totalcount[chrom]["raw"] = 0
                    totalcount[chrom]["ref"] = 0

                for rtype in statistics[chrom]:
                    if not rtype in readnum:
                        readnum[rtype] = 0
                    for end in statistics[chrom][rtype]:
                        totalcount[chrom][rtype] += statistics[chrom][rtype][end]
                        readnum[rtype] += statistics[chrom][rtype][end]
                        allends[chrom][rtype][end] = statistics[chrom][rtype][end]
                        for x in range(-5, 0, 1):
                            newend = end[x:]
                            if newend in allends[chrom][rtype]:
                                allends[chrom][rtype][newend] += statistics[chrom][
                                    rtype
                                ][end]
                            else:
                                allends[chrom][rtype][newend] = statistics[chrom][
                                    rtype
                                ][end]

            outstr = []
            for chrom in allends:
                for rtype in allends[chrom]:
                    for end in allends[chrom][rtype]:
                        if allends[chrom][rtype][end] is not "":
                            tosave = None
                            if fastadict:
                                info = ""
                                for id in fastadict:
                                    if (id.upper() == chrom.upper()) or any(
                                        [
                                            x.upper() in chrom.upper().split(".")[-1]
                                            for x in fastadict[id]
                                        ]
                                    ):
                                        info = fastadict[id]
                                tosave = str.join(
                                    "\t",
                                    [
                                        str(chrom),
                                        rtype,
                                        end,
                                        str(allends[chrom][rtype][end]),
                                        str(
                                            allends[chrom][rtype][end]
                                            / totalcount[chrom][rtype]
                                        ),
                                        str(
                                            allends[chrom][rtype][end] / readnum[rtype]
                                        ),
                                        str(",".join(info)),
                                    ],
                                )
                            else:
                                tosave = str.join(
                                    "\t",
                                    [
                                        str(chrom),
                                        rtype,
                                        end,
                                        str(allends[chrom][rtype][end]),
                                        str(
                                            allends[chrom][rtype][end]
                                            / totalcount[chrom][rtype]
                                        ),
                                        str(
                                            allends[chrom][rtype][end] / readnum[rtype]
                                        ),
                                    ],
                                )
                            if tosave and len(tosave) > 1:
                                outstr.append(tosave)

            with gzip.open(out, "ab") as o:
                o.write(bytes("\n".join(outstr), encoding="UTF-8"))

        else:
            with gzip.open(out, "wb") as o:
                o.write(bytes("", encoding="UTF-8"))

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


def printlog(msg):
    with open("log", "a") as l:
        print(str(msg), file=l)


if __name__ == "__main__":
    args = parseargs()
    collectends(
        args.bams,
        args.outdir,
        args.procs,
        args.verbosity,
        args.offset,
        args.fasta,
        args.lookup,
    )

#
# CountEnds.py ends here



# Logger.py ---
#
# Filename: Logger.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Mon Aug 12 10:26:55 2019 (+0200)
# Version:
# Package-Requires: ()
# Last-Updated: Wed Apr 29 16:42:40 2020 (+0200)
#           By: Joerg Fallmann
#     Update #: 91
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:
import logging
import os
import sys
import traceback as tb


def makelogdir(logdir):
    if not os.path.isabs(logdir):
        logdir = os.path.abspath(logdir)
    if not os.path.exists(logdir):
        os.makedirs(logdir)
    return logdir


def setup_logger(
    name,
    log_file,
    filemode="w",
    logformat=None,
    datefmt=None,
    level="WARNING",
    delay=False,
):
    """Function setup as many loggers as you want"""

    logger = logging.getLogger(name)
    if log_file != "stdout" and log_file != "stderr":
        makelogdir(os.path.dirname(log_file))
        if not os.path.isfile(os.path.abspath(log_file)):
            open(os.path.abspath(log_file), "a").close()
        handler = logging.FileHandler(
            os.path.abspath(log_file), mode=filemode, delay=delay
        )
    else:
        handler = logging.StreamHandler()

    handler.setFormatter(logging.Formatter(fmt=logformat, datefmt=datefmt))

    logger.setLevel(level)
    logger.addHandler(handler)

    return logger


if __name__ == "__main__":
    try:
        # set up logging to file
        log = setup_logger(
            name="",
            log_file="stderr",
            logformat="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
            datefmt="%m-%d %H:%M",
            level="WARNING",
        )

    except Exception:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        logging.error("".join(tbe.format()))


# Logger.py ends here



# Cluster2tRNA.py ---
#
# Filename: Cluster2tRNA.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Tue Oct  9 15:18:52 2018 (+0200)
# Version:
# Package-Requires: ()
# Last-Updated: Tue Oct  9 21:36:52 2018 (+0200)
#           By: Joerg Fallmann
#     Update #: 40
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:

### IMPORTS
import argparse
from io import StringIO
import os
import gzip
import re
import sys
import string
import collections
import traceback as tb
import inspect

####load own modules
cmd_subfolder = os.path.join(
    os.path.dirname(
        os.path.realpath(os.path.abspath(inspect.getfile(inspect.currentframe())))
    ),
    "../../MONSDA",
)
if cmd_subfolder not in sys.path:
    sys.path.insert(0, cmd_subfolder)
# sys.path=[str(os.getenv('HOME')+"/Work/Scripts/Python/lib")] + sys.path
from Collection import *

####Biopython stuff
from Bio import SeqIO
from Bio.Seq import Seq

####numpy and matplolib and pyplot
# import numpy as np
# import matplotlib
# import matplotlib.pyplot as plt
from Logger import *

### MAIN
def parseargs():
    parser = argparse.ArgumentParser(
        description="Read BAM file read by read and collect mapping statistics"
    )
    parser.add_argument(
        "-f", "--fasta", type=str, help="Reference genome FASTA, needs to be indexed"
    )
    parser.add_argument(
        "-o", "--outdir", type=str, default=".", help="Output directory"
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        type=int,
        default=0,
        choices=[0, 1],
        help="increase output verbosity",
    )

    return parser.parse_args()


if __name__ == "__main__":
    try:
        args = parseargs()
        if args.outdir:
            print("Checking or creating outdir " + str(args.outdir))
            if not os.path.isabs(args.outdir):
                outdir = os.path.abspath(args.outdir)
            if not os.path.exists(args.outdir):
                os.makedirs(args.outdir)
        else:
            outdir = os.path.abspath(os.getcwd())

        translater = cluster2trna(args.fasta)
        outstring = []

        for cluster in translater["cluster"]:

            outstring.append(
                "\t".join([cluster, ",".join(translater["cluster"][cluster])])
            )

        for chrom in translater["tRNA"]:
            for strand in translater["tRNA"][chrom]:
                for tRNA in translater["tRNA"][chrom][strand]:
                    outstring.append("\t".join([tRNA, chrom, strand]))

        with open(os.path.join(outdir, "cluster2trna_" + args.fasta), "w") as o:
            print("\n".join(outstring), file=o)

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        with open("error", "a") as h:
            print("".join(tbe.format()), file=h)


#
# Cluster2tRNA.py ends here



#!/usr/bin/env python
# Code from:https://github.com/vivekbhr/Subread_to_DEXSeq

import sys
import collections
import itertools
import os.path
import optparse
import os
import gzip
import warnings
import traceback as tb
import logging
import inspect

cmd_subfolder = os.path.join(
    os.path.dirname(
        os.path.realpath(os.path.abspath(inspect.getfile(inspect.currentframe())))
    ),
    "../../lib",
)
if cmd_subfolder not in sys.path:
    sys.path.insert(0, cmd_subfolder)

from Logger import *

try:
    scriptname = os.path.basename(inspect.stack()[-1].filename).replace(".py", "")
    log = logging.getLogger(scriptname)

    lvl = log.level if log.level else "INFO"
    for handler in log.handlers[:]:
        handler.close()
        log.removeHandler(handler)

    try:
        handler = logging.FileHandler("LOGS/MONSDA.log", mode="a")
    except:
        handler = logging.FileHandler("log", mode="a")
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    log.setLevel(lvl)

except Exception:
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = tb.TracebackException(
        exc_type,
        exc_value,
        exc_tb,
    )
    print("".join(tbe.format()), file=sys.stderr)


optParser = optparse.OptionParser(
    usage="python %prog [options] <in.gtf> <out.gff>",
    description="Script to prepare annotation for DEXSeq."
    + "This script takes an annotation file in Ensembl GTF format"
    + "and outputs a 'flattened' annotation file suitable for use "
    + "with the count_in_exons.py script ",
    epilog="Written by Simon Anders (sanders@fs.tum.de), European Molecular Biology "
    + "Laboratory (EMBL). (c) 2010. Released under the terms of the GNU General "
    + "Public License v3. Part of the 'DEXSeq' package. "
    + "Modified by Vivek Bhardwaj (just a bit!) to write featurecounts gtf as an option.",
)

optParser.add_option(
    "-r",
    "--aggregate",
    type="choice",
    dest="aggregate",
    choices=("no", "yes"),
    default="yes",
    help="'yes' or 'no'. Indicates whether two or more genes sharing an exon should be merged into an 'aggregate gene'. If 'no', the exons that can not be assiged to a single gene are ignored.",
)

# add option for featurecounts output
optParser.add_option(
    "-f",
    "--featurecountsgtf",
    type="string",
    dest="fcgtf",
    action="store",
    help="gtf file to write for featurecounts.",
)
optParser.add_option(
    "-s",
    "--stranded",
    dest="fcstrand",
    action="store_true",
    default=False,
    help="Use strand in gffparse if stranded sequencing protocol was used.",
)
optParser.add_option(
    "--nextflow", dest="nextflow", action="store_true", help="Run in nextflow mode"
)
##

(opts, args) = optParser.parse_args()
print(opts, args)

if len(args) != 2:
    sys.stderr.write("Script to prepare annotation for DEXSeq.\n\n")
    sys.stderr.write(
        "Usage: python %s <in.gtf> <out.gff>\n\n" % os.path.basename(sys.argv[0])
    )
    sys.stderr.write("This script takes an annotation file in Ensembl GTF format\n")
    sys.stderr.write("and outputs a 'flattened' annotation file suitable for use\n")
    sys.stderr.write("with the count_in_exons.py script.\n")
    sys.exit(1)

try:
    import HTSeq
except ImportError:
    sys.stderr.write(
        "Could not import HTSeq. Please install the HTSeq Python framework\n"
    )
    sys.stderr.write("available from http://www-huber.embl.de/users/anders/HTSeq\n")
    sys.exit(1)


gtf_file = args[0]
out_file = args[1]

aggregateGenes = opts.aggregate == "yes"

# Step 1: Store all exons with their gene and transcript ID
# in a GenomicArrayOfSets
try:
    exons = HTSeq.GenomicArrayOfSets("auto", stranded=opts.fcstrand)
except:
    warnings.warn(
        "There was an issue setting stranded to True when creating the HTSeq.GenomicArrayOfSets. This is probably due to an error in the annotation allocating one gene ID to both strands. Will retry without strand information, please carefully check your results."
    )
    exons = HTSeq.GenomicArrayOfSets("auto", stranded=False)

for f in HTSeq.GFF_Reader(gtf_file):
    if f.type != "exon":
        continue
    f.attr["gene_id"] = f.attr["gene_id"].replace(":", "_")
    exons[f.iv] += (f.attr["gene_id"], f.attr["transcript_id"])


# Step 2: Form sets of overlapping genes

# We produce the dict 'gene_sets', whose values are sets of gene IDs. Each set
# contains IDs of genes that overlap, i.e., share bases (on the same strand).
# The keys of 'gene_sets' are the IDs of all genes, and each key refers to
# the set that contains the gene.
# Each gene set forms an 'aggregate gene'.

if aggregateGenes == True:
    gene_sets = collections.defaultdict(lambda: set())
    for iv, s in exons.steps():
        # For each step, make a set, 'full_set' of all the gene IDs occuring
        # in the present step, and also add all those gene IDs, whch have been
        # seen earlier to co-occur with each of the currently present gene IDs.
        full_set = set()
        for gene_id, transcript_id in s:
            full_set.add(gene_id)
            full_set |= gene_sets[gene_id]
        # Make sure that all genes that are now in full_set get associated
        # with full_set, i.e., get to know about their new partners
        for gene_id in full_set:
            assert gene_sets[gene_id] <= full_set
            gene_sets[gene_id] = full_set


# Step 3: Go through the steps again to get the exonic sections. Each step
# becomes an 'exonic part'. The exonic part is associated with an
# aggregate gene, i.e., a gene set as determined in the previous step,
# and a transcript set, containing all transcripts that occur in the step.
# The results are stored in the dict 'aggregates', which contains, for each
# aggregate ID, a list of all its exonic_part features.

aggregates = collections.defaultdict(lambda: list())
for iv, s in exons.steps():
    # Skip empty steps
    if len(s) == 0:
        continue
    gene_id = list(s)[0][0]
    ## if aggregateGenes=FALSE, ignore the exons associated to more than one gene ID
    if aggregateGenes == False:
        check_set = set()
        for geneID, transcript_id in s:
            check_set.add(geneID)
        if len(check_set) > 1:
            continue
        else:
            aggregate_id = gene_id
    # Take one of the gene IDs, find the others via gene sets, and
    # form the aggregate ID from all of them
    else:
        assert set(gene_id for gene_id, transcript_id in s) <= gene_sets[gene_id]
        aggregate_id = "+".join(gene_sets[gene_id])
    # Make the feature and store it in 'aggregates'
    f = HTSeq.GenomicFeature(aggregate_id, "exonic_part", iv)
    f.source = os.path.basename(sys.argv[0])
    #   f.source = "camara"
    f.attr = {}
    f.attr["gene_id"] = aggregate_id
    transcript_set = set((transcript_id for gene_id, transcript_id in s))
    f.attr["transcripts"] = "+".join(transcript_set)
    aggregates[aggregate_id].append(f)


# Step 4: For each aggregate, number the exonic parts

aggregate_features = []
for l in aggregates.values():
    for i in range(len(l) - 1):
        if l[i].iv.strand != l[i + 1].iv.strand:
            raise ValueError(
                "Same name found on two strands: %s, %s" % (str(l[i]), str(l[i + 1]))
            )
        assert l[i].name == l[i + 1].name, str(l[i + 1]) + " has wrong name"
        assert l[i].iv.end <= l[i + 1].iv.start, str(l[i + 1]) + " starts too early"
        if l[i].iv.chrom != l[i + 1].iv.chrom:
            raise ValueError(
                "Same name found on two chromosomes: %s, %s"
                % (str(l[i]), str(l[i + 1]))
            )
    aggr_feat = HTSeq.GenomicFeature(
        l[0].name,
        "aggregate_gene",
        HTSeq.GenomicInterval(
            l[0].iv.chrom, l[0].iv.start, l[-1].iv.end, l[0].iv.strand
        ),
    )
    aggr_feat.source = os.path.basename(sys.argv[0])
    aggr_feat.attr = {"gene_id": aggr_feat.name}
    for i in range(len(l)):
        l[i].attr["exonic_part_number"] = "%03d" % (i + 1)
    aggregate_features.append(aggr_feat)


# Step 5: Sort the aggregates, then write everything out

aggregate_features.sort(key=lambda f: (f.iv.chrom, f.iv.start))

if ".gz" in out_file:
    with gzip.open(out_file, "wb") as fout:
        for aggr_feat in aggregate_features:
            fout.write(bytes(aggr_feat.get_gff_line(), encoding="UTF8"))
            for f in aggregates[aggr_feat.name]:
                fout.write(bytes(f.get_gff_line(), encoding="UTF8"))

else:
    with open(out_file, "w") as fout:
        for aggr_feat in aggregate_features:
            fout.write(aggr_feat.get_gff_line())
            for f in aggregates[aggr_feat.name]:
                fout.write(f.get_gff_line())

## modify file to print gtf if featurecounts gtf requested
fcountgtf = opts.fcgtf

if fcountgtf:
    if ".gz" in fcountgtf:
        os.system(
            "zcat "
            + out_file
            + "|sed s/aggregate_gene/gene/g|sed s/exonic_part/exon/g |cut -d';' -f1|gzip > "
            + fcountgtf
        )
    else:
        os.system(
            "sed s/aggregate_gene/gene/g "
            + out_file
            + "|sed s/exonic_part/exon/g |cut -d';' -f1 > "
            + fcountgtf
        )



#!/usr/bin/env python3
# RemoveSoftClip.py ---
#
# Filename: RemoveSoftClip.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Tue May 15 12:44:06 2018 (+0200)
# Version:
# Package-Requires: ()
# Last-Updated: Wed Jun 20 11:44:19 2018 (+0200)
#           By: Joerg Fallmann
#     Update #: 320
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:

### IMPORTS
import argparse
import pprint  # pp = pprint.PrettyPrinter(indent=4) #pp.pprint(stuff)
from io import StringIO
import os
import gzip
import re
import sys
import pysam
from pyfaidx import Fasta
import traceback
import logging
import traceback as tb
import inspect

# Container
import collections

### MAIN

cmd_subfolder = os.path.join(
    os.path.dirname(
        os.path.realpath(os.path.abspath(inspect.getfile(inspect.currentframe())))
    ),
    "../lib",
)
if cmd_subfolder not in sys.path:
    sys.path.insert(0, cmd_subfolder)

from Logger import *

try:
    scriptname = os.path.basename(inspect.stack()[-1].filename).replace(".py", "")
    log = logging.getLogger(scriptname)

    lvl = log.level if log.level else "INFO"
    for handler in log.handlers[:]:
        handler.close()
        log.removeHandler(handler)

    try:
        handler = logging.FileHandler("LOGS/MONSDA.log", mode="a")
    except:
        handler = logging.FileHandler("log", mode="a")
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    log.setLevel(lvl)

except Exception:
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = tb.TracebackException(
        exc_type,
        exc_value,
        exc_tb,
    )
    print("".join(tbe.format()), file=sys.stderr)


def parseargs():
    parser = argparse.ArgumentParser(
        description="Read BAM file read by read and remove softclips"
    )
    parser.add_argument(
        "-f", "--fasta", type=str, help="Reference genome FASTA, needs to be indexed"
    )
    parser.add_argument(
        "-b",
        "--bams",
        type=str,
        help="Mapped reads BAM(s) comma separated, need to be sorted and indexed",
    )
    parser.add_argument("-o", "--out", type=str, default=None, help="Output file name")
    parser.add_argument(
        "-c",
        "--cluster",
        default=False,
        action="store_true",
        help="Parses cluster information from chromosome tag",
    )
    parser.add_argument(
        "--loglevel",
        type=str,
        default="INFO",
        choices=["WARNING", "ERROR", "INFO", "DEBUG"],
        help="Set log level",
    )
    return parser.parse_args()


def process(fasta, bams, outname, cluster=None):
    outdir = os.path.dirname(outname) if outname else None
    if outdir:
        print("Checking or creating outdir " + str(outdir))
        if not os.path.isabs(outdir):
            outdir = os.path.abspath(outdir)
        if not os.path.exists(outdir):
            os.makedirs(outdir)
    else:
        outdir = os.path.abspath(os.getcwd())

    for bam in bams.split(","):
        fn = os.path.basename(bam)
        out = os.path.join(outdir, fn + "_nosoftclip.bam") if not outname else outname

        if os.path.isfile(out):
            print("File " + out + " exists, will be deleted")
            os.remove(out)

        samfile = parse_bam(bam)

        try:
            samfile.check_index()
        except:
            print("No index for file: " + bam + ". Please create!")

        remove_clip(bam, fasta, out, cluster)


def remove_clip(bam, fasta, out, cluster=None):
    samfile = parse_bam(bam)
    header = samfile.header

    # if cluster:  # SQ tags need to be edited
    #    for k, v in header['SQ'].items():
    #        s, t, n, chrom, coord = k.split(':')
    #        k = chrom

    with pysam.Samfile(out, "wb", template=samfile) as bamout:
        for read in samfile.fetch():
            # newread = pysam.AlignedSegment(header)
            # if not read.is_unmapped:
            chrom = read.reference_name
            mate_chrom = read.next_reference_name
            start, end, strand = (0, 0, ".")

            if cluster and "::" in chrom:  # Hammerhead_1::SM_V7_1:2251747-2251831(+)
                t, n, chrom, coord = chrom.split(":")
                start, end = map(int, coord.split("(")[0].split("-"))
                strand = coord.split("(")[1].replace(")", "")
                if mate_chrom:
                    t, n, mate_chrom, mcoord = (
                        mate_chrom.split(":")
                        if ":" in mate_chrom
                        else (None, None, mate_chrom, None)
                    )

            cigar = read.cigarstring
            if not cigar or "*" in cigar:
                bamout.write(read)
                continue

            newcigar = []
            clip_5 = 0
            clip_3 = 0

            changed = False
            inseq = False
            for op, length in read.cigar:
                if op == 5:  # H
                    changed = True
                elif op == 4:  # S
                    changed = True
                    if not inseq:
                        clip_5 = length
                    else:
                        clip_3 = length
                else:
                    inseq = True
                    newcigar.append((op, length))

            qual = read.query_alignment_qualities
            read.query_sequence = read.query_alignment_sequence
            try:
                read.query_qualities = qual
            except:
                log.info("Alignment quality not available " + str(read.query_qualities))

            """
            read.reference_start does not need to be changed, according to SAM specs
            S for softclip does NOT consume a reference base and is thus not considered 
            for Alignment start position!
            https://samtools.github.io/hts-specs/SAMv1.pdf
            EXCEPT if we have cluster enabled!
            """
            if cluster:
                read.reference_start = read.reference_start + start

            read.cigar = newcigar
            read.reference_name = chrom
            read.next_reference_name = mate_chrom

            bamout.write(read)

    close_bam(samfile)


def parse_bam(bam):
    if ".bam" in bam:
        return pysam.AlignmentFile(bam, "rb")
    elif ".sam" in bam:
        return pysam.AlignmentFile(bam, "r")


def read_head(bam):
    if ".bam" in bam:
        return pysam.AlignmentFile(bam, "rb").header
    elif ".sam" in bam:
        return pysam.AlignmentFile(bam, "r").header


def close_bam(samfile):
    samfile.close()


def check_idx(file):
    if (
        (os.path.isfile(file + ".idx"))
        or (os.path.isfile(str.join(".", split(".", file)[0:-1], "fai")))
        or (os.path.isfile(str.join(".", split(".", file)[0:-1], "faidx")))
    ):
        return True
    else:
        return False


###MAIN###

if __name__ == "__main__":
    args = parseargs()
    if args.loglevel:
        log.setLevel(args.loglevel)
    process(args.fasta, args.bams, args.out, args.cluster)

#
# RemoveSoftClip.py ends here



#!/usr/bin/env python3
import sys
import os
import re
import math
import argparse
import subprocess
from snakemake.utils import read_job_properties


##############################
# Helper functions
##############################
def _get_default_partition():
    """Retrieve default partition for cluster"""
    if "":
        return ""
    cmd = "sinfo -O \"partition\""
    res = subprocess.run(cmd, check=True, shell=True,
                         stdout=subprocess.PIPE)
    m = re.search("(?P<partition>\S+)\*", res.stdout.decode(), re.M)
    partition = m.group("partition")
    return partition


def _get_cluster_configuration(partition):
    """Retrieve cluster configuration for a partition."""
    # Retrieve partition info; we tacitly assume we only get one response
    cmd = " ".join(
        ["sinfo -e -O \"partition,cpus,memory,time,size,maxcpuspernode\"",
         "-h -p {}".format(partition)])
    res = subprocess.run(cmd, check=True, shell=True, stdout=subprocess.PIPE)
    m = re.search("(?P<partition>\S+)\s+(?P<cpus>\d+)\s+(?P<memory>\S+)\s+((?P<days>\d+)-)?(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)\s+(?P<size>\S+)\s+(?P<maxcpus>\S+)",
                  res.stdout.decode())
    if m is None:
        m = re.search("(?P<partition>\S+)\s+(?P<cpus>\d+)\s+(?P<memory>\S+)\s+(?P<time>\S+)\s+(?P<size>\S+)\s+(?P<maxcpus>\S+)",
                      res.stdout.decode())
    d = m.groupdict()
    if not 'days' in d or not d['days'] or 'time' in d and d['time'] == 'infinite':
        d['days'] = 30
        d['hours'] = d['minutes'] = d['seconds'] = 0
    d["time"] = int(d['days']) * 24 * 60 + \
        int(d['hours']) * 60 + int(d['minutes']) + \
        math.ceil(int(d['seconds']) / 60)
    return d


def _get_features_and_memory(partition):
    """Retrieve features and memory for a partition in the cluster
    configuration. """
    cmd = " ".join(
        ["sinfo -e -O \"memory,features_act\"",
         "-h -p {}".format(partition)])
    res = subprocess.run(cmd, check=True, shell=True, stdout=subprocess.PIPE)
    mem_feat = []
    for x in res.stdout.decode().split("\n"):
        if not re.search("^\d+", x):
            continue
        m = re.search("^(?P<mem>\d+)\s+(?P<feat>\S+)", x)
        mem_feat.append({'mem': m.groupdict()["mem"],
                         'features': m.groupdict()["feat"].split(",")})
    return mem_feat


def _get_available_memory(mem_feat, constraints=None):
    """Get available memory

    If constraints are given, parse constraint string into array of
    constraints and compare them to active features. Currently only
    handles comma-separated strings and not the more advanced
    constructs described in the slurm manual.

    Else, the minimum memory for a given partition is returned.

    """
    if constraints is None:
        return min([int(x['mem']) for x in mem_feat])
    try:
        constraint_set = set(constraints.split(","))
        for x in mem_feat:
            if constraint_set.intersection(x["features"]) == constraint_set:
                return int(x["mem"])
    except Exception as e:
        print(e)
        raise


def _update_memory_and_ntasks(arg_dict, MEMORY_PER_CPU, MEMORY_PER_PARTITION):
    """Given a one-node job, update memory and ntasks if the requested
    configuration is unavailable"""
    if arg_dict["mem"] is not None:
        arg_dict["mem"] = min(int(arg_dict["mem"]),
                              MEMORY_PER_PARTITION)
        AVAILABLE_MEM = arg_dict["ntasks"] * MEMORY_PER_CPU
        if arg_dict["mem"] > AVAILABLE_MEM:
            arg_dict["ntasks"] = int(math.ceil(arg_dict["mem"] /
                                               MEMORY_PER_CPU))
    arg_dict["ntasks"] = min(int(config["cpus"]),
                             int(arg_dict["ntasks"]))

##############################
# Main script
##############################
parser = argparse.ArgumentParser(add_help=False)
parser.add_argument(
    "--help", help="Display help message.", action="store_true")
parser.add_argument(
    "positional", action="append",
    nargs="?", metavar="POS",
    help="additional arguments not in slurm parser group to pass to sbatch")

# A subset of SLURM-specific arguments. Note that the parser is used
# implicitly in that the parsed arguments are used to construct a
# dictionary of arguments to be processed. Specific arguments can be
# modified via the cluster configuration file.
slurm_parser = parser.add_argument_group("slurm-specific arguments")
slurm_parser.add_argument(
    "-a", "--array", help="job array index values")
slurm_parser.add_argument(
    "-A", "--account", help="charge job to specified account")
slurm_parser.add_argument(
    "--begin", help="defer job until HH:MM MM/DD/YY")
slurm_parser.add_argument(
    "-c", "--cpus-per-task", help="number of cpus required per task",
    type=int, default=1)
slurm_parser.add_argument(
    "-d", "--dependency",
    help="defer job until condition on jobid is satisfied")
slurm_parser.add_argument(
    "-D", "--workdir", help="set working directory for batch script")
slurm_parser.add_argument(
    "-e", "--error", help="file for batch script's standard error",
    default="" if "" else None)
slurm_parser.add_argument(
    "-J", "--job-name", help="name of job")
slurm_parser.add_argument(
    "--mail-type", help="notify on state change: BEGIN, END, FAIL or ALL")
slurm_parser.add_argument(
    "--mail-user", help="who to send email notification for job state changes")
slurm_parser.add_argument(
    "-n", "--ntasks", help="number of tasks to run", type=int, default=1)
slurm_parser.add_argument(
    "-N", "--nodes", help="number of nodes on which to run (N = min[-max])",
    type=int)
slurm_parser.add_argument(
    "-o", "--output", help="file for batch script's standard output",
    default="" if "" else None)
slurm_parser.add_argument(
    "-p", "--partition", help="partition requested",
    default=_get_default_partition(), type=str)
slurm_parser.add_argument(
    "-q", "--qos", help="quality of service")
slurm_parser.add_argument(
    "-Q", "--quiet", help="quiet mode (suppress informational messages)")
slurm_parser.add_argument(
    "-t", "--time", help="time limit")
slurm_parser.add_argument(
    "--wrap", help="wrap command string in a sh script and submit")
slurm_parser.add_argument(
    "-C", "--constraint", help="specify a list of constraints")
slurm_parser.add_argument(
    "--mem", help="minimum amount of real memory")

opt_keys = ["array", "account", "begin", "cpus_per_task",
            "dependency", "workdir", "error", "job_name", "mail_type",
            "mail_user", "ntasks", "nodes", "output", "partition",
            "quiet", "time", "wrap", "constraint", "mem"]

args = parser.parse_args()

if args.help:
    parser.print_help()
    sys.exit(0)

jobscript = sys.argv[-1]
job_properties = read_job_properties(jobscript)

extras = ""
if args.positional:
    for m in args.positional:
        if m is not None:
            extras = extras + " " + m

arg_dict = dict(args.__dict__)

# Set default account
if arg_dict["account"] is None:
    if "" != "":
        arg_dict["account"] = ""


# Ensure output folder for Slurm log files exist.
# This is a bit hacky; will run for every Slurm submission...
if arg_dict["output"] is not None:
    if not os.path.exists(os.path.dirname(arg_dict["output"])):
        os.makedirs(os.path.dirname(arg_dict["output"]))
if arg_dict["error"] is not None:
    if not os.path.exists(os.path.dirname(arg_dict["error"])):
        os.makedirs(os.path.dirname(arg_dict["error"]))


# Process resources
if "resources" in job_properties:
    resources = job_properties["resources"]
    if arg_dict["time"] is None:
        if "runtime" in resources:
            arg_dict["time"] = resources["runtime"]
        elif "walltime" in resources:
            arg_dict["time"] = resources["walltime"]
    if arg_dict["mem"] is None:
        if "mem" in resources:
            arg_dict["mem"] = resources["mem"]
        elif "mem_mb" in resources:
            arg_dict["mem"] = resources["mem_mb"]

# Threads
if "threads" in job_properties:
    arg_dict["ntasks"] = job_properties["threads"]


# Process cluster configuration. Note that setting time, ntasks, and
# mem will override any setting in resources. It is assumed that the
# cluster configuration parameters are named according to the sbatch
# long options names
cluster_config = job_properties.get("cluster", {})
arg_dict.update(job_properties.get("cluster", {}))

# Determine partition with features. If no constraints have been set,
# select the partition with lowest memory
try:
    part = arg_dict["partition"]
    config = _get_cluster_configuration(part)
    mem_feat = _get_features_and_memory(part)
    MEMORY_PER_PARTITION = _get_available_memory(mem_feat,
                                                 arg_dict["constraint"])
    MEMORY_PER_CPU = MEMORY_PER_PARTITION / int(config["cpus"])
except subprocess.CalledProcessError as e:
    print(e)
    raise e
except Exception as e:
    print(e)
    raise e


# Update time. If requested time is larger than maximum allowed time,
# reset
try:
    if arg_dict["time"] is not None:
        arg_dict["time"] = min(int(config["time"]), int(arg_dict["time"]))
except Exception as e:
    print(e)
    raise e


# Adjust memory in the single-node case only; getting the
# functionality right for multi-node multi-cpu jobs requires more
# development
if arg_dict["nodes"] is None or int(arg_dict["nodes"]) == 1:
    _update_memory_and_ntasks(arg_dict, MEMORY_PER_CPU,
                              MEMORY_PER_PARTITION)
else:
    if int(arg_dict["ntasks"]) == 1:
        # Allocate at least as many tasks as requested nodes
        arg_dict["ntasks"] = int(arg_dict["nodes"])


opts = ""
for k, v in arg_dict.items():
    if k not in opt_keys:
        continue
    if v is not None:
        opts += " --{} \"{}\" ".format(k.replace("_", "-"), v)

if arg_dict["wrap"] is not None:
    cmd = "sbatch {opts}".format(opts=opts)
else:
    cmd = "sbatch {opts} {extras}".format(opts=opts, extras=extras)

try:
    res = subprocess.run(cmd, check=True, shell=True, stdout=subprocess.PIPE)
except subprocess.CalledProcessError as e:
    raise e

# Get jobid
res = res.stdout.decode()
try:
    m = re.search("Submitted batch job (\d+)", res)
    jobid = m.group(1)
    print(jobid)
except Exception as e:
    print(e)
    raise



#!/usr/bin/env python3
"""
Snakemake SLURM submit script.
"""
import warnings  # use warnings.warn() rather than print() to output info in this script

from snakemake.utils import read_job_properties

import slurm_utils

# cookiecutter arguments
SBATCH_DEFAULTS = """ """
CLUSTER_CONFIG = "cluster_config.yaml"
ADVANCED_ARGUMENT_CONVERSION = {"yes": True, "no": False}[
    "no"
]

RESOURCE_MAPPING = {
    "time": ("time", "runtime", "walltime"),
    "mem": ("mem", "mem_mb", "ram", "memory"),
    "mem-per-cpu": ("mem-per-cpu", "mem_per_cpu", "mem_per_thread"),
    "nodes": ("nodes", "nnodes"),
}

# parse job
jobscript = slurm_utils.parse_jobscript()
job_properties = read_job_properties(jobscript)

sbatch_options = {}
cluster_config = slurm_utils.load_cluster_config(CLUSTER_CONFIG)

# 1) sbatch default arguments
sbatch_options.update(slurm_utils.parse_sbatch_defaults(SBATCH_DEFAULTS))

# 2) cluster_config defaults
sbatch_options.update(cluster_config["__default__"])

# 3) Convert resources (no unit conversion!) and threads
sbatch_options.update(
    slurm_utils.convert_job_properties(job_properties, RESOURCE_MAPPING)
)

# 4) cluster_config for particular rule
sbatch_options.update(cluster_config.get(job_properties.get("rule"), {}))

# 5) cluster_config options
sbatch_options.update(job_properties.get("cluster", {}))

# 6) Advanced conversion of parameters
if ADVANCED_ARGUMENT_CONVERSION:
    sbatch_options = slurm_utils.advanced_argument_conversion(sbatch_options)

# 7) Format pattern in snakemake style
sbatch_options = slurm_utils.format_values(sbatch_options, job_properties)

# ensure sbatch output dirs exist
for o in ("output", "error"):
    slurm_utils.ensure_dirs_exist(sbatch_options[o]) if o in sbatch_options else None

# submit job and echo id back to Snakemake (must be the only stdout)
print(slurm_utils.submit_job(jobscript, **sbatch_options))



#! /usr/bin/env python3

import sys
import argparse
import os
import inspect
import gzip
import glob
import re
import logging
import traceback as tb
from collections import defaultdict


cmd_subfolder = os.path.join(
    os.path.dirname(
        os.path.realpath(os.path.abspath(inspect.getfile(inspect.currentframe())))
    ),
    "../lib",
)
if cmd_subfolder not in sys.path:
    sys.path.insert(0, cmd_subfolder)

from Logger import *

try:
    scriptname = os.path.basename(inspect.stack()[-1].filename).replace(".py", "")
    log = logging.getLogger(scriptname)

    lvl = log.level if log.level else "INFO"
    for handler in log.handlers[:]:
        handler.close()
        log.removeHandler(handler)

    try:
        handler = logging.FileHandler("LOGS/MONSDA.log", mode="a")
        handler.setFormatter(
            logging.Formatter(
                fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
                datefmt="%m-%d %H:%M",
            )
        )
        log.addHandler(handler)
    except:  # This is not working for nextflow so we just log to a dummy log file and nextflow takes care of the rest
        handler = logging.FileHandler("log", mode="a")
        handler.setFormatter(
            logging.Formatter(
                fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
                datefmt="%m-%d %H:%M",
            )
        )
        log.addHandler(handler)

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    log.setLevel(lvl)

except Exception:
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = tb.TracebackException(
        exc_type,
        exc_value,
        exc_tb,
    )
    print("".join(tbe.format()), file=sys.stderr)


def parseargs():
    parser = argparse.ArgumentParser(
        description="Create Rmd snippets for Summary from downstream analyses results"
    )
    parser.add_argument(
        "--files",
        dest="files",
        required=True,
        type=str,
        nargs="*",
        help="Names of input files",
    )
    parser.add_argument(
        "--output", dest="output", required=True, type=str, help="output Rmd File"
    )
    parser.add_argument(
        "--env", dest="env", required=True, type=str, help="used conda environment"
    )
    parser.add_argument("--loglevel", default="INFO", help="Log verbosity")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


# Class
class NestedDefaultDict(defaultdict):
    def __init__(self, *args, **kwargs):
        super(NestedDefaultDict, self).__init__(NestedDefaultDict, *args, **kwargs)


def makeoutdir(outdir):
    if not os.path.isabs(outdir):
        outdir = os.path.abspath(outdir)
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    return outdir


def pretty(d, indent=0):
    for key, value in d.items():
        log.debug("\t" * indent + str(key))
        if isinstance(value, dict):
            pretty(value, indent + 1)
        else:
            log.debug("\t" * (indent + 1) + str(value))


# Analyses
compile = {
    "DE": "Differential Gene Expression",
    "DEU": "Differential Exon Usage",
    "DAS": "Differential Alternative Splicing",
    "DTU": "Differential Transcript Usage",
}


def check_workflow(files):
    logid = scriptname + ".check_workflow: "
    files = [os.path.basename(file) for file in files]
    works = set()
    for file in files:
        log.debug(logid + str(file))
        w = file.split("_")[0].replace("[", "").replace("]", "")
        if w not in ["Sig", "SigDOWN", "SigUP"]:
            works.add(w)
    log.debug(logid + str(works))
    if len(works) > 1:
        log.error(
            logid
            + "several Workflows found: > "
            + str(works)
            + " < check for wrong file naming."
        )
        sys.exit(1)
    try:
        compile[list(works)[0]]
    except:
        log.error(logid + "workflow > " + str(works) + " < not found")
        sys.exit(1)
    log.debug(logid + "workflow found: " + str(works))
    return list(works)[0]


def create_file_tree(files):
    logid = scriptname + ".create_file_tree: "
    tree = NestedDefaultDict()
    files = [
        file
        for file in files
        if not any(x in file for x in ["noshrink", "AllConditions", "DataSet"])
        and any(x in file for x in ["results", "figure_"])
    ]
    for file in files:
        setting = os.path.basename(file).split(".", 1)[0]
        setting = (
            setting.replace("Sig_", "").replace("SigUP_", "").replace("SigDOWN_", "")
        )
        if "SESSION" in file:
            WF, TOOL, COMBI, NAME = setting.split("_")
            tree[WF][TOOL][COMBI]["DataSet"][NAME] = file
            continue
        filename = setting.split("_")
        if len(filename) < 7:
            WF, TOOL, COMBI, COMP, RES, NAME = setting.split("_")
            EXT = os.path.basename(file).split(".", 1)[1]
            tree[WF][TOOL][COMBI][COMP][RES][NAME] = file
        else:
            continue
    log.debug(logid + "FILE-TREE: " + str(pretty(tree)))
    return tree


def integrate_table(file):
    chunk = f"""
```{{r, table:{file}, include = TRUE, echo = FALSE, message = TRUE, warning = FALSE}}
library(DT)
#table <- read.table(gzfile(paste(params$root,'{file}',sep='/')), header = TRUE)
table <- read.table(gzfile('{file}'), header = TRUE)
DT::datatable(table)
```

"""
    return chunk


def integrate_figures(files, islist=False):
    listlines = list()
    img = list()
    listlines.append(
        f"```{{r, echo=FALSE, out.width='{100/len(files)}%', out.height='{100/len(files)}%',fig.cap='',fig.show='hold',fig.align='center'}}"
    )
    counter = 1
    for file in files:
        f = file.replace("\n", "")
        if islist == True:
            listlines.append(f"fig.{counter} <- paste(params$root,'{f}', sep='/')")
        else:
            listlines.append(f"fig.{counter} <- '{f}'")
        img.append(f"path.expand(fig.{counter})")
        counter += 1
    listlines.append(f"knitr::include_graphics(c({','.join(img)}), rel_path=FALSE)")
    listlines.append("```\n\n")
    listlines.append(f"\n***\n")
    return "\n".join(listlines)


def integrate_list(file):
    listlines = list()
    elements = NestedDefaultDict()
    with open(file, "r") as listfile:
        for line in listfile:
            if "geneID" in line:
                continue
            id = line.split("\t")[0]
            name = line.split("\t")[1]
            file = line.split("\t")[2]
            elements[id][name][file]
    for id in elements.keys():
        listlines.append(f"**{id} : {list(elements[id].keys())[0]}** \n")
        for name in elements[id].keys():
            listlines.append(integrate_figures(list(elements[id][name]), islist=True))
    return "\n".join(listlines)


def create_Rmd(files, output, env):
    logid = scriptname + ".create_Rmd: "
    outdir = os.path.dirname(output)
    makeoutdir(outdir)
    log.debug(f"{logid} FILES: {files}")
    workflow = check_workflow(files)
    tree = create_file_tree(files)
    lines = list()

    # Add title
    lines.append(f"# {compile[workflow]} {{.tabset}} \n\n")

    # Create file structure
    for tool in tree[workflow].keys():
        lines.append(f"## {tool}  {{.tabset}} \n\n")

        for combi in tree[workflow][tool].keys():
            lines.append(f"### {combi}  {{.tabset}} \n\n")

            for comparison in tree[workflow][tool][combi].keys():
                lines.append(f"#### {comparison}  {{.tabset}} \n\n")

                if "figure" in tree[workflow][tool][combi][comparison].keys():
                    lines.append(f"##### FIGURES  {{.tabset}} \n\n")
                    lines.append(f"###### OVERVIEW \n\n")
                    for name in tree[workflow][tool][combi][comparison][
                        "figure"
                    ].keys():
                        lines.append(f"\n<br />\n")
                        lines.append(f"\n**{name}** \n")
                        lines.append(
                            f"\n```\n{tree[workflow][tool][combi][comparison]['figure'][name]}\n```\n"
                        )
                        lines.append(
                            integrate_figures(
                                [
                                    tree[workflow][tool][combi][comparison]["figure"][
                                        name
                                    ]
                                ]
                            )
                        )

                if "list" in tree[workflow][tool][combi][comparison].keys():
                    for name in tree[workflow][tool][combi][comparison]["list"].keys():
                        lines.append(f"\n##### {name} \n")
                        lines.append(f"\n<br />\n")
                        lines.append(
                            integrate_list(
                                tree[workflow][tool][combi][comparison]["list"][name]
                            )
                        )

                if "table" in tree[workflow][tool][combi][comparison].keys():
                    lines.append(f"##### TABLES  \n")
                    for name in tree[workflow][tool][combi][comparison]["table"].keys():
                        lines.append(f"\n<br />\n")
                        lines.append(f"\n**{name}**  \n")
                        # lines.append(integrate_table(tree[workflow][tool][combi][comparison]["table"][name]["tsv.gz"]))
                        lines.append(
                            f"\n```\n{tree[workflow][tool][combi][comparison]['table'][name]}\n```\n"
                        )

                if "SESSION" in tree[workflow][tool][combi][comparison].keys():
                    lines.append(f"##### R-SESSION  {{.tabset}} \n\n")
                    lines.append(f"\n<br />\n")
                    lines.append(
                        f"To access the {tool} R-session, follow these steps\n\n"
                    )
                    lines.append(
                        f"1) Create the corresponding conda environment and activate it:  \n"
                    )
                    lines.append(f"```\nconda env create -f {env}.yaml\n```  \n")
                    lines.append(f"2) Start R and load the workspace:  \n")
                    lines.append(
                        f"```\nload('{tree[workflow][tool][combi][comparison]['SESSION']}')\n```\n"
                    )

    log.debug(logid + "lines: " + str(lines))

    if os.path.exists(output):
        os.rename(output, output + ".bak")
    with open(output, "a") as writefile:
        for line in lines:
            writefile.write(line)
        writefile.write("\n\n")


####################
####    MAIN    ####
####################

if __name__ == "__main__":
    logid = scriptname + ".main: "
    try:
        args = parseargs()
        if args.loglevel:
            log.setLevel(args.loglevel)

        log.debug(logid + str(log.handlers))

        create_Rmd(args.files, args.output, args.env)

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        log.error(logid + "".join(tbe.format()))


# RmdCreator.py ends here



# logger.py ---
#
# Filename: logger.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Mon Aug 12 10:26:55 2019 (+0200)
# Version:
# Package-Requires: ()
# Last-Updated: Tue Sep 24 16:53:41 2019 (+0200)
#           By: Joerg Fallmann
#     Update #: 64
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:
import logging
import multiprocessing
import os
import sys
import inspect
import traceback as tb


def makelogdir(logdir):
    if not os.path.isabs(logdir):
        logdir = os.path.abspath(logdir)
    if not os.path.exists(logdir):
        os.makedirs(logdir)
    return logdir


def setup_logger(
    name, log_file, filemode="w", logformat=None, datefmt=None, level="WARNING"
):
    """Function setup as many loggers as you want"""

    logger = logging.getLogger(name)
    if log_file != "stdout" and log_file != "stderr":
        makelogdir(os.path.dirname(log_file))
        handler = logging.FileHandler(log_file, mode=filemode)
    else:
        handler = logging.StreamHandler()

    handler.setFormatter(logging.Formatter(fmt=logformat, datefmt=datefmt))

    logger.setLevel(level)
    logger.addHandler(handler)

    return logger


if __name__ == "__main__":
    try:
        # set up logging to file
        logging = setup_logger(
            name="",
            log_file="stderr",
            logformat="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
            datefmt="%m-%d %H:%M",
            level="WARNING",
        )

        # define a Handler which writes INFO messages or higher to the sys.stderr
        # console = logging.StreamHandler()
        # console.setLevel(logging.INFO)
        # set a format which is simpler for console use
        # formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
        # tell the handler to use this format
        # console.setFormatter(formatter)
        # add the handler to the root logger
        # logging.getLogger('').addHandler(console)

        # Now, we can log to the root logger, or any other logger. First the root...
        # logging.info('Imported logger.py')
        # Now, use this in code defining a couple of other loggers which might represent areas in your
        # application, e.g.:
        # log = logging.getLogger('logger.main')

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type,
            exc_value,
            exc_tb,
        )
        logging.error("".join(tbe.format()))


# log.py ends here



#!/usr/bin/env python3
import os
import sys
from os.path import dirname
import re
import math
import argparse
import subprocess as sp
from io import StringIO

from snakemake import io
from snakemake.io import Wildcards
from snakemake.utils import SequenceFormatter
from snakemake.utils import AlwaysQuotedFormatter
from snakemake.utils import QuotedFormatter
from snakemake.exceptions import WorkflowError
from snakemake.logging import logger


def parse_jobscript():
    """Minimal CLI to require/only accept single positional argument."""
    p = argparse.ArgumentParser(description="SLURM snakemake submit script")
    p.add_argument("jobscript", help="Snakemake jobscript with job properties.")
    return p.parse_args().jobscript


def parse_sbatch_defaults(parsed):
    """Unpack SBATCH_DEFAULTS."""
    d = parsed.split() if type(parsed) == str else parsed
    args = {}
    for keyval in [a.split("=") for a in d]:
        k = keyval[0].strip().strip("-")
        v = keyval[1].strip() if len(keyval) == 2 else None
        args[k] = v
    return args


def load_cluster_config(path):
    """Load config to dict

    Load configuration to dict either from absolute path or relative
    to profile dir.
    """
    if path:
        path = os.path.join(dirname(__file__), os.path.expandvars(path))
        dcc = io.load_configfile(path)
    else:
        dcc = {}
    if "__default__" not in dcc:
        dcc["__default__"] = {}
    return dcc


# adapted from format function in snakemake.utils
def format(_pattern, _quote_all=False, **kwargs):  # noqa: A001
    """Format a pattern in Snakemake style.
    This means that keywords embedded in braces are replaced by any variable
    values that are available in the current namespace.
    """
    fmt = SequenceFormatter(separator=" ")
    if _quote_all:
        fmt.element_formatter = AlwaysQuotedFormatter()
    else:
        fmt.element_formatter = QuotedFormatter()
    try:
        return fmt.format(_pattern, **kwargs)
    except KeyError as ex:
        raise NameError(
            f"The name {ex} is unknown in this context. Please "
            "make sure that you defined that variable. "
            "Also note that braces not used for variable access "
            "have to be escaped by repeating them "
        )


#  adapted from Job.format_wildcards in snakemake.jobs
def format_wildcards(string, job_properties):
    """ Format a string with variables from the job. """

    class Job(object):
        def __init__(self, job_properties):
            for key in job_properties:
                setattr(self, key, job_properties[key])

    job = Job(job_properties)
    if "params" in job_properties:
        job._format_params = Wildcards(fromdict=job_properties["params"])
    else:
        job._format_params = None
    if "wildcards" in job_properties:
        job._format_wildcards = Wildcards(fromdict=job_properties["wildcards"])
    else:
        job._format_wildcards = None
    _variables = dict()
    _variables.update(
        dict(params=job._format_params, wildcards=job._format_wildcards)
    )
    if hasattr(job, "rule"):
        _variables.update(dict(rule=job.rule))
    try:
        return format(string, **_variables)
    except NameError as ex:
        raise WorkflowError(
            "NameError with group job {}: {}".format(job.jobid, str(ex))
        )
    except IndexError as ex:
        raise WorkflowError(
            "IndexError with group job {}: {}".format(job.jobid, str(ex))
        )


# adapted from ClusterExecutor.cluster_params function in snakemake.executor
def format_values(dictionary, job_properties):
    formatted = dictionary.copy()
    for key, value in list(formatted.items()):
        if isinstance(value, str):
            try:
                formatted[key] = format_wildcards(value, job_properties)
            except NameError as e:
                msg = "Failed to format cluster config " "entry for job {}.".format(
                    job_properties["rule"]
                )
                raise WorkflowError(msg, e)
    return formatted


def convert_job_properties(job_properties, resource_mapping=None):
    options = {}
    if resource_mapping is None:
        resource_mapping = {}
    resources = job_properties.get("resources", {})
    for k, v in resource_mapping.items():
        options.update({k: resources[i] for i in v if i in resources})

    if "threads" in job_properties:
        options["cpus-per-task"] = job_properties["threads"]
    return options


def ensure_dirs_exist(path):
    """Ensure output folder for Slurm log files exist."""
    di = dirname(path)
    if di == "":
        return
    if not os.path.exists(di):
        os.makedirs(di, exist_ok=True)
    return


def format_sbatch_options(**sbatch_options):
    """Format sbatch options"""
    options = []
    for k, v in sbatch_options.items():
        val = ""
        if v is not None:
            val = f"={v}"
        options.append(f"--{k}{val}")
    return options


def submit_job(jobscript, **sbatch_options):
    """Submit jobscript and return jobid."""
    options = format_sbatch_options(**sbatch_options)
    try:
        cmd = ["sbatch"] + ["--parsable"] + options + [jobscript]
        res = sp.check_output(cmd)
    except sp.CalledProcessError as e:
        raise e
    # Get jobid
    res = res.decode()
    try:
        jobid = re.search(r"(\d+)", res).group(1)
    except Exception as e:
        raise e
    return jobid


def advanced_argument_conversion(arg_dict):
    """Experimental adjustment of sbatch arguments to the given or default partition."""
    # Currently not adjusting for multiple node jobs
    nodes = int(arg_dict.get("nodes", 1))
    if nodes > 1:
        return arg_dict
    partition = arg_dict.get("partition", None) or _get_default_partition()
    constraint = arg_dict.get("constraint", None)
    ncpus = int(arg_dict.get("cpus-per-task", 1))
    runtime = arg_dict.get("time", None)
    config = _get_cluster_configuration(partition, constraint, arg_dict.get("mem", 0))
    mem = arg_dict.get("mem", ncpus * min(config["MEMORY_PER_CPU"]))
    if mem > max(config["MEMORY"]):
        logger.info(
            f"requested memory ({mem}) > max memory ({max(config['MEMORY'])}); "
            "adjusting memory settings"
        )
        mem = max(config["MEMORY"])

    # Calculate available memory as defined by the number of requested
    # cpus times memory per cpu
    AVAILABLE_MEM = ncpus * min(config["MEMORY_PER_CPU"])
    # Add additional cpus if memory is larger than AVAILABLE_MEM
    if mem > AVAILABLE_MEM:
        logger.info(
            f"requested memory ({mem}) > "
            f"ncpus x MEMORY_PER_CPU ({AVAILABLE_MEM}); "
            "trying to adjust number of cpus up"
        )
        ncpus = int(math.ceil(mem / min(config["MEMORY_PER_CPU"])))
    if ncpus > max(config["CPUS"]):
        logger.info(
            f"ncpus ({ncpus}) > available cpus ({max(config['CPUS'])}); "
            "adjusting number of cpus down"
        )
        ncpus = min(int(max(config["CPUS"])), ncpus)
    adjusted_args = {"mem": int(mem), "cpus-per-task": ncpus}

    # Update time. If requested time is larger than maximum allowed time, reset
    if runtime:
        runtime = time_to_minutes(runtime)
        time_limit = max(config["TIMELIMIT_MINUTES"])
        if runtime > time_limit:
            logger.info(
                f"time (runtime) > time limit {time_limit}; " "adjusting time down"
            )
            adjusted_args["time"] = time_limit

    # update and return
    arg_dict.update(adjusted_args)
    return arg_dict


timeformats = [
    re.compile(r"^(?P<days>\d+)-(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)$"),
    re.compile(r"^(?P<days>\d+)-(?P<hours>\d+):(?P<minutes>\d+)$"),
    re.compile(r"^(?P<days>\d+)-(?P<hours>\d+)$"),
    re.compile(r"^(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)$"),
    re.compile(r"^(?P<minutes>\d+):(?P<seconds>\d+)$"),
    re.compile(r"^(?P<minutes>\d+)$"),
]


def time_to_minutes(time):
    """Convert time string to minutes.

    According to slurm:

      Acceptable time formats include "minutes", "minutes:seconds",
      "hours:minutes:seconds", "days-hours", "days-hours:minutes"
      and "days-hours:minutes:seconds".

    """
    if not isinstance(time, str):
        time = str(time)
    d = {"days": 0, "hours": 0, "minutes": 0, "seconds": 0}
    regex = list(filter(lambda regex: regex.match(time) is not None, timeformats))
    if len(regex) == 0:
        return
    assert len(regex) == 1, "multiple time formats match"
    m = regex[0].match(time)
    d.update(m.groupdict())
    minutes = (
        int(d["days"]) * 24 * 60
        + int(d["hours"]) * 60
        + int(d["minutes"])
        + math.ceil(int(d["seconds"]) / 60)
    )
    assert minutes > 0, "minutes has to be greater than 0"
    return minutes


def _get_default_partition():
    """Retrieve default partition for cluster"""
    res = sp.check_output(["sinfo", "-O", "partition"])
    m = re.search(r"(?P<partition>\S+)\*", res.decode(), re.M)
    partition = m.group("partition")
    return partition


def _get_cluster_configuration(partition, constraints=None, memory=0):
    """Retrieve cluster configuration.

    Retrieve cluster configuration for a partition filtered by
    constraints, memory and cpus

    """
    try:
        import pandas as pd
    except ImportError:
        print(
            "Error: currently advanced argument conversion "
            "depends on 'pandas'.", file=sys.stderr
        )
        sys.exit(1)

    if constraints:
        constraint_set = set(constraints.split(","))
    cmd = ["sinfo", "-e", "-o", "%all", "-p", partition]
    try:
        output = sp.Popen(" ".join(cmd), shell=True, stdout=sp.PIPE).communicate()
    except Exception as e:
        print(e)
        raise
    data = re.sub(" \\|", "|", output[0].decode())
    df = pd.read_csv(StringIO(data), sep="|")
    try:
        df["TIMELIMIT_MINUTES"] = df["TIMELIMIT"].apply(time_to_minutes)
        df["MEMORY_PER_CPU"] = df["MEMORY"] / df["CPUS"]
        df["FEATURE_SET"] = df["AVAIL_FEATURES"].str.split(",").apply(set)
    except Exception as e:
        print(e)
        raise
    if constraints:
        constraint_set = set(constraints.split(","))
        i = df["FEATURE_SET"].apply(lambda x: len(x.intersection(constraint_set)) > 0)
        df = df.loc[i]
    memory = min(memory, max(df["MEMORY"]))
    df = df.loc[df["MEMORY"] >= memory]
    return df



#/usr/bin/env python3
# AddStructure.py ---
#
# Filename: AddStructure.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Tue Sep 10 18:00:42 2019 (+0200)
# Version:
# Package-Requires: ()
# Last-Updated: Wed Sep 11 09:10:01 2019 (+0200)
#           By: Joerg Fallmann
#     Update #: 33
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:
###Imports
import sys,os
import argparse
import traceback as tb
import gzip
import RNA

###Arguments
def parseargs():
	parser = argparse.ArgumentParser(description='Add structure to (bed) file containing sequence')
	parser.add_argument("-f", "--field", type=int, default=0, help='Which field contains the sequence, default is last field (0)')
	parser.add_argument("-b", "--bed", type=str, help='Bed or other tab separated file containing the sequence')

	return parser.parse_args()

###CODE

def addseq(field, bed):
    try:
        entries = parse_bed(bed)
        for line in entries:
            sequence = line.rstrip().split('\t')[field-1].upper()
            # create model details
            md = RNA.md()
            # create new fold_compound object
            fc = RNA.fold_compound(sequence, md)
            # compute minimum free energy (mfe) and corresponding structure
            (ss, mfe) = fc.mfe()

            sys.stdout.write('\t'.join([line.strip(),ss])+'\n')

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type, exc_value, exc_tb,
        )
        with open('error','a') as h:
            print(''.join(tbe.format()), file=h)

def parse_bed(bed, annotated=None):
    try:
        if os.path.isfile(os.path.abspath(bed)):
            if '.gz' in bed:
                return gzip.open(bed,'rt')
            else:
                return open(bed,'rt')

    except Exception as err:
        exc_type, exc_value, exc_tb = sys.exc_info()
        tbe = tb.TracebackException(
            exc_type, exc_value, exc_tb,
        )
        with open('error','a') as h:
            print(''.join(tbe.format()), file=h)


###MAIN
if __name__ == '__main__':
	args=parseargs()
	addseq(args.field, args.bed)

#
# AddStructure.py ends here



# Utils.py ---
#
# Filename: Utils.py
# Description:
# Author: Joerg Fallmann
# Maintainer:
# Created: Tue Sep 18 15:39:06 2018 (+0200)
# Version:
# Package-Requires: ()
# Last-Updated: Thu Feb  4 18:01:07 2021 (+0100)
#           By: Joerg Fallmann
#     Update #: 2888
# URL:
# Doc URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#

# Change Log:
#
#
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
#
#

# Code:
# import os, sys, inspect
# # realpath() will make your script run, even if you symlink it :)
# cmd_folder = os.path.dirname(os.path.realpath(os.path.abspath( inspect.getfile( inspect.currentframe() )) ))
# if cmd_folder not in sys.path:
#     sys.path.insert(0, cmd_folder)

#
# # Use this if you want to include modules from a subfolder
# cmd_subfolder = os.path.join(os.path.dirname(os.path.realpath(os.path.abspath( inspect.getfile( inspect.currentframe() )) )),"lib")
# if cmd_subfolder not in sys.path:
#     sys.path.insert(0, cmd_subfolder)
#
# # Info:
# # cmd_folder = os.path.dirname(os.path.abspath(__file__)) # DO NOT USE __file__ !!!
# # __file__ fails if the script is called in different ways on Windows.
# # __file__ fails if someone does os.chdir() before.
# # sys.argv[0] also fails, because it doesn't not always contain the path.

import collections
import datetime
import functools
import glob
import gzip
import hashlib
import heapq
import inspect
import itertools
import json
import logging
import os
import re
import shutil
import subprocess
import sys
import traceback as tb
from collections import OrderedDict, defaultdict
from io import StringIO
from operator import itemgetter

import numpy as np
import six
from Bio import SeqIO
from natsort import natsorted
from snakemake.common.configfile import load_configfile

try:
    scriptname = (
        os.path.basename(inspect.stack()[-1].filename)
        .replace("Run", "")
        .replace(".py", "")
    )
    log = logging.getLogger(scriptname)

    lvl = log.level if log.level else "INFO"
    for handler in log.handlers[:]:
        handler.close()
        log.removeHandler(handler)
    handler = logging.FileHandler("LOGS/MONSDA.log", mode="a")
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(name)-12s %(message)s",
            datefmt="%m-%d %H:%M",
        )
    )
    log.addHandler(handler)
    log.setLevel(lvl)

except Exception:
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = tb.TracebackException(
        exc_type,
        exc_value,
        exc_tb,
    )
    print("".join(tbe.format()), file=sys.stderr)


#  NestedDefaultDict
class NestedDefaultDict(defaultdict):
    def __init__(self, *args, **kwargs):
        super(NestedDefaultDict, self).__init__(NestedDefaultDict, *args, **kwargs)

    def __repr__(self):
        return repr(dict(self))

    def merge(self, *args):
        self = merge_dicts(self, *args)


# Wrapper
def check_run(func):
    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        logid = scriptname + ".Collection_func_wrapper: "
        try:
            return func(*args, **kwargs)

        except Exception:
            exc_type, exc_value, exc_tb = sys.exc_info()
            tbe = tb.TracebackException(
                exc_type,
                exc_value,
                exc_tb,
            )
            log.error(logid + "".join(tbe.format()))

    return func_wrapper


@check_run
def rmempty(check):
    ret = list()
    for f in check:
        if os.path.isfile(f):
            ret.append(f)
    return ret


@check_run
def comment_remover(
    textlist,
):  # inspired by https://stackoverflow.com/questions/241327/remove-c-and-c-comments-using-python
    def replacer(match):
        s = match.group(0)
        if s.startswith("/"):
            pass  # Do nothing
        else:
            return s

    pattern = re.compile(
        r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
        re.DOTALL | re.MULTILINE,
    )
    return [re.sub(pattern, replacer, x) for x in textlist]


##############################
#########Python Subs##########
##############################
@check_run
def dict_inst(d):
    logid = scriptname + ".Collection_dict_inst: "
    if (
        isinstance(d, dict)
        or isinstance(d, OrderedDict)
        or isinstance(d, defaultdict)
        or isinstance(d, NestedDefaultDict)
    ):
        return True


@check_run
def get_from_dict(dataDict, mapList):
    logid = scriptname + ".Collection_get_from_dict: "
    log.debug(logid + "MAPLIST: " + str(mapList) + "\tDict: " + str(dataDict))
    ret = dataDict
    for k in mapList:
        log.debug(logid + "k: " + str(k))
        if dataDict.get(k):
            dataDict = dataDict[k]
            log.debug(logid + "sub_dict: " + str(dataDict))
        else:
            return list([])
    if ret != dataDict:
        log.debug(logid + "RET: " + str(dataDict))
        return list([dataDict])
    else:
        log.debug(logid + "RET: " + str(list([])))
        return list([])


@check_run
def yield_from_dict(key, dictionary):
    for k, v in dictionary.items():
        if k == key:
            yield v
        elif dict_inst(v):
            for result in yield_from_dict(key, v):
                yield result
        elif isinstance(v, list):
            for d in v:
                if dict_inst(d):
                    for result in yield_from_dict(key, d):
                        yield result


@check_run
def sub_dict(dataDict, mapList):
    logid = scriptname + ".Collection_sub_dict: "
    log.debug(logid + str(mapList))
    ret = dataDict
    for k in mapList:
        log.debug(logid + "k: " + str(k))
        if k in ret:
            ret = ret[k]
            if not dict_inst(ret):
                ret = {k: ret}
        else:
            log.debug(logid + "No k in dict")
            return dict()
    return ret


@check_run
def subset_dict(dataDict, mapList):
    logid = scriptname + ".Collection_subset_dict: "
    log.debug(logid + str(mapList))
    parse = sub_dict(dataDict, mapList)
    ret = {}
    nested_set(ret, mapList, parse)
    log.debug(logid + str(ret))
    return ret


@check_run
def nested_set(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value


@check_run
def merge_dicts(d, u):
    # https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    # python 3.8+ compatibility
    try:
        collectionsAbc = collections.abc
    except:
        collectionsAbc = collections

    for k, v in six.iteritems(u):
        dv = d.get(k, {})
        if not isinstance(dv, collectionsAbc.Mapping):
            d[k] = v
        elif isinstance(v, collectionsAbc.Mapping):
            d[k] = merge_dicts(dv, v)
        else:
            d[k] = v
    return d


@check_run
def keysets_from_dict(
    dictionary, search=None, original=None
):  # Only works for equal depth keysets, needs tweaking for other use cases
    logid = scriptname + ".Collection_keysets_from_dict: "

    keylist = list()
    if dict_inst(dictionary):
        for k, v in keys_from_dict(dictionary, search).items():
            keylist.append(v)
        log.debug(logid + "kl:" + str(keylist))
        combis = list()
        for i in range(1, len(keylist) + 1):
            subkeylist = keylist[0:i]
            combis.extend(list(itertools.product(*subkeylist)))
        log.debug(logid + "cs:" + str(combis))
        ret = list()
        for combi in combis:
            check = sub_dict(dictionary, combi)
            log.debug(logid + "checking: " + str(check))
            if (
                isvalid(check)
                and (isinstance(check, dict) and check.get("SAMPLES"))
                or isinstance(check, str)
            ):
                log.debug(logid + "found: " + str(combi))
                ret.append(combi)
        return ret
    else:
        return keylist


@check_run
def keys_from_dict(dictionary, search=None, save=None, first=True, lvl=0):
    logid = scriptname + ".Collection_keys_from_dict: "

    if first:
        first = False
        end = depth(dictionary)
        save = defaultdict(list)
        log.debug(logid + "TOTALDEPTH:" + str(end))

    if dict_inst(dictionary):
        log.debug(logid + "dictDEPTH: " + str(depth(dictionary)))
        log.debug(logid + "dictLEVEL: " + str(lvl))
        for k, v in dictionary.items():
            if not search or (search and k != search):
                save[lvl].append(k)
                log.debug(logid + "TMPSAVE: " + str(save))
                if dict_inst(v):
                    save = keys_from_dict(v, search, save, first, lvl + 1)
                else:
                    continue
            else:
                log.debug(logid + "Found search: " + str(save))
                return save
        return save
    else:
        return save


@check_run
def depth(d):
    if dict_inst(d):
        return 1 + (max(map(depth, d.values())) if d else 0)
    return 0


@check_run
def list_all_keys_of_dict(dictionary):
    logid = scriptname + ".Collection_list_all_keys_of_dict: "
    for key, value in dictionary.items():
        if type(value) is dict:
            yield key
            yield from list_all_keys_of_dict(value)
        else:
            yield key


@check_run
def list_all_values_of_dict(dictionary):
    if dict_inst(dictionary):
        for key, value in dictionary.items():
            if dict_inst(value):
                yield from list_all_values_of_dict(value)
            else:
                yield (key, value)
    else:
        yield dictionary


@check_run
def find_all_values_on_key(key, dictionary):
    if dict_inst(dictionary):
        for k, v in dictionary.items():
            if dict_inst(v):
                yield from find_all_values_on_key(key, v)
            elif k == key:
                yield v

    else:
        return dictionary


@check_run
def find_key_for_value(val, dictionary):
    logid = scriptname + ".Collection_find_key_for_value: "
    log.debug(logid + "VAL: " + str(val) + " Dict: " + str(dictionary))
    if dict_inst(dictionary):
        for k, v in dictionary.items():
            log.debug(logid + "DICT: " + str(k) + " ; " + str(v))
            if dict_inst(v):
                log.debug(logid + "item" + str(v))
                yield from find_key_for_value(val, v)
            elif v == val or val in v:
                yield k
    else:
        return dictionary


@check_run
def value_extract(key, var):
    logid = scriptname + ".Collection_value_extract: "
    log.debug(logid + str(var))
    if hasattr(var, "items"):
        for k, v in var.items():
            if k == key:
                yield v
            if dict_inst(v):
                for result in value_extract(key, v):
                    yield result
            elif isinstance(v, list):
                for d in v:
                    for result in value_extract(key, d):
                        yield result


@check_run
def find_innermost_value_from_dict(dictionary):
    logid = scriptname + ".Collection_find_innermost_value_from_dict: "
    if dict_inst(dictionary):
        for k, v in dictionary.items():
            if dict_inst(v):
                return find_innermost_value_from_dict(v)
            else:
                return v
    else:
        return dictionary


@check_run
def removekey(d, key):
    logid = scriptname + ".Collection_removekey: "
    r = dict(d)
    del r[key]
    return r


@check_run
def getlowest_list(a, n):
    if n > len(a) - 1:
        b = len(a) - 1
    else:
        b = n
    if len(a) > 0 and n > 0:
        return list(np.partition(a, b)[:n])
    else:
        return list(None for i in range(n))


@check_run
def gethighest_list(a, n):
    if len(a) - n < 0:
        b = len(a) - 1
    else:
        b = len(a) - n
    if len(a) > 0 and n > 0:
        return list(np.partition(a, b)[-n:])
    else:
        return list(None for i in range(n))


@check_run
def getlowest_dict(a, n):
    if n > len(a):
        b = len(a)
    else:
        b = n
    if len(a) > 0:
        return dict(heapq.nsmallest(b, a.items(), key=itemgetter(1)))
    else:
        return dict({i: None for i in range(n)})


@check_run
def gethighest_dict(a, n):
    if n > len(a):
        b = len(a)
    else:
        b = n
    if len(a) > 0:
        return dict(heapq.nlargest(b, a.items(), key=itemgetter(1)))
    else:
        return dict({i: None for i in range(n)})


@check_run
def toarray(file, ulim):
    x = np.loadtxt(
        str(file),
        usecols=(ulim),
        delimiter="\t",
        unpack=True,
        converters={ulim: lambda s: convertcol(s.decode("utf-8"))},
    )
    return x


@check_run
def convertcol(entry):
    if isinvalid(entry):
        return np.nan
    else:
        return float(entry)


@check_run
def isvalid(x=None):
    if x:
        if x in ("None", "nan", "none", "NA", "NAN") or x is None or x is np.nan:
            return False
        else:
            return True
    else:
        return False


@check_run
def isinvalid(x=None):
    if x:
        if x in ("None", "nan", "none", "NA", "NAN") or x is None or x is np.nan:
            return True
        else:
            return False
    else:
        return True


@check_run
def makeoutdir(outdir):
    if not os.path.isabs(outdir):
        outdir = os.path.abspath(outdir)
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    return outdir


@check_run
def parseseq(sequence):
    if isinstance(sequence, StringIO):
        seq = sequence

    elif isinstance(sequence, str) and os.path.isfile(sequence):
        if ".gz" in sequence:
            seq = gzip.open(sequence, "rt")
        else:
            seq = open(sequence, "rt")
    else:
        header = ">Seq1:default:nochrom:(.)"
        s = sequence
        seq = StringIO("{header}\n{s}".format(header=header, s=s))

    return seq


@check_run
def npprint(a, o=None):  # , format_string ='{0:.2f}'):
    out = ""
    it = np.nditer(a, flags=["f_index"])
    while not it.finished:
        out += "%d\t%0.7f" % (it.index + 1, it[0]) + "\n"
        it.iternext()
    if o:
        o.write(bytes(out, encoding="UTF-8"))
    else:
        print(out)


@check_run
def idfromfa(id):
    goi, chrom, strand = [None, None, None]
    try:
        goi, chrom = id.split(":")[::2]
        strand = str(id.split(":")[3].split("(")[1][0])
    except:
        print(
            "Fasta header is not in expected format, you will loose information on strand and chromosome"
        )
        goi = id
        chrom, strand = ["na", "na"]

    if goi and chrom and strand:
        return [str(goi), str(chrom), str(strand)]
    else:
        sys.exit(
            "Could not assign any value from fasta header, please check your fasta files"
        )


@check_run
def cluster2trna(seqs):
    translater = collections.OrderedDict()
    translater["cluster"] = collections.OrderedDict()
    translater["tRNA"] = collections.OrderedDict()

    for fa in SeqIO.parse(seqs, "fasta"):
        head = str(
            fa.id
        ).upper()  # cluster1:chr19.tRNA5-LysCTT(+) cluster2:NC_007091.3.tRNA25-ArgTCT
        cluster, info = head.split(":")
        chrom, trna = (info.split(".")[0], info.split(".")[-1].split("(")[0])
        strand = "u"
        if "(+)" in info or "(-)" in info:
            strand = re.sub("[()]", "_", info.split(".")[-1]).split("_")[1]

        if cluster in translater["cluster"]:
            translater["cluster"][cluster].append(trna)
        else:
            translater["cluster"][cluster] = list()
            translater["cluster"][cluster].append(trna)
        if chrom in translater["tRNA"]:
            if strand in translater["tRNA"][chrom]:
                translater["tRNA"][chrom][strand].append(trna)
            else:
                translater["tRNA"][chrom][strand] = list()
                translater["tRNA"][chrom][strand].append(trna)
        else:
            translater["tRNA"][chrom] = collections.OrderedDict()
            translater["tRNA"][chrom][strand] = list()
            translater["tRNA"][chrom][strand].append(trna)

    return translater


@check_run
def check_ref(reference):
    if os.path.exists(os.path.abspath(reference)):
        return reference
    elif os.path.exists(os.path.abspath(reference + ".gz")):
        return reference + ".gz"


@check_run
def multi_replace(repl, text):
    print("MULTI: " + str(repl) + str(text))
    # Create a regular expression  from the dictionary keys
    regex = re.compile("(%s)" % "|".join(map(re.escape, repl.keys())))

    # For each match, look-up corresponding value in dictionary
    return regex.sub(lambda mo: dict[mo.string[mo.start() : mo.end()]], text)


@check_run
def makelogdir(logdir):
    if not os.path.isabs(logdir):
        logdir = os.path.abspath(logdir)
    if not os.path.exists(logdir):
        os.makedirs(logdir)
    return logdir


@check_run
def get_dict_hash(d):
    logid = scriptname + ".get_dict_hash: "
    log.debug(logid + "INPUT DICT: " + str(d))
    ret = str(hashlib.sha256(bytes(str(sorted(d.items())), "utf-8")).hexdigest())
    log.debug(logid + "HASH: " + ret)
    return ret


@check_run
def add_to_innermost_key_by_list(addto, toadd, keylist):
    logid = scriptname + ".add_to_innermost_key_by_list: "
    log.debug(logid + str(addto) + ", " + str(toadd))

    tconf = {}
    for i in range(
        len(keylist)
    ):  # need to add options as last element to dict of unknown depth
        tconf[keylist[i]] = {}
        tconf = tconf[keylist[i]]
        if i == len(keylist) - 1:
            tconf = toadd

    addto.update(tconf)
    return addto


#
# Utils.py ends here
